import gzip
import os
import shutil

from datasets import DatasetDict, Split

from zshot.config import MODELS_CACHE_PATH
from zshot.evaluation.dataset.dataset import DatasetWithEntities
from zshot.evaluation.dataset.med_mentions.entities import MEDMENTIONS_ENTITIES, MEDMENTIONS_SPLITS, \
    MEDMENTIONS_TYPE_INV
from zshot.evaluation.dataset.med_mentions.utils import preprocess_medmentions
from zshot.utils import download_file

LABELS = MEDMENTIONS_ENTITIES

FILES = [
    "corpus_pubtator.txt",
    "corpus_pubtator.txt.gz",
    "corpus_pubtator_pmids_all.txt",
    "corpus_pubtator_pmids_dev.txt",
    "corpus_pubtator_pmids_test.txt",
    "corpus_pubtator_pmids_train.txt"
]


def _unzip(file):
    with gzip.open(file, 'rb') as f_in:
        with open(file.replace(".gz", ""), 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)


def _download_raw_data(path):
    txt_files = [
        "https://raw.githubusercontent.com/chanzuckerberg/MedMentions/master/full/data/corpus_pubtator_pmids_all.txt",
        "https://raw.githubusercontent.com/chanzuckerberg/MedMentions/master/full/data/corpus_pubtator_pmids_dev.txt",
        "https://raw.githubusercontent.com/chanzuckerberg/MedMentions/master/full/data/corpus_pubtator_pmids_test.txt",
        "https://raw.githubusercontent.com/chanzuckerberg/MedMentions/master/full/data/corpus_pubtator_pmids_trng.txt"
    ]
    for file in txt_files:
        download_file(file, path)
    shutil.move(os.path.join(path, "corpus_pubtator_pmids_trng.txt"),
                os.path.join(path, "corpus_pubtator_pmids_train.txt"))
    gz_file = "https://raw.githubusercontent.com/chanzuckerberg/MedMentions/master/st21pv/data/corpus_pubtator.txt.gz"
    download_file(gz_file, path)
    zip_file = os.path.join(path, "corpus_pubtator.txt.gz")
    _unzip(zip_file)


def _delete_temporal_files(cache_path):
    for file in FILES:
        os.remove(os.path.join(cache_path, file))


def _create_split_dataset(data, split):
    dataset = DatasetWithEntities.from_dict(
        {
            "tokens": [[tok.word for tok in sentence] for sentence in data],
            "ner_tags": [[tok.label_id for tok in sentence] for sentence in data]
        },
        split=split,
        entities=list(
            filter(lambda ent: MEDMENTIONS_TYPE_INV[ent.name] in MEDMENTIONS_SPLITS[str(split)],
                   MEDMENTIONS_ENTITIES))
    )
    return dataset


def load_medmentions() -> DatasetDict[DatasetWithEntities]:
    _download_raw_data(MODELS_CACHE_PATH)
    train_sentences, dev_sentences, test_sentences = preprocess_medmentions(MODELS_CACHE_PATH)
    _delete_temporal_files(MODELS_CACHE_PATH)

    medmentions_zs = DatasetDict()
    for split, sentences in [(Split.TRAIN, train_sentences),
                             (Split.VALIDATION, dev_sentences),
                             (Split.TEST, test_sentences)]:
        medmentions_zs[split] = _create_split_dataset(sentences, split)

    return medmentions_zs



import argparse
import os
import pkgutil
from enum import Enum
from typing import List, Iterator, Optional, Union

from spacy.tokens import Doc

from zshot.config import MODELS_CACHE_PATH
from zshot.linker.linker import Linker
from zshot.utils.data_models import Span
from zshot.utils.file_utils import download_file

BLINK_ENTITIES = "http://dl.fbaipublicfiles.com/BLINK/entity.jsonl"

BLINK_ENTITIES_ENCODING = "http://dl.fbaipublicfiles.com/BLINK/all_entities_large.t7"

BLINK_BI_ENCODER_FILES = \
    ["http://dl.fbaipublicfiles.com/BLINK/biencoder_wiki_large.json",
     "http://dl.fbaipublicfiles.com/BLINK/biencoder_wiki_large.bin"]
BLINK_CROSS_ENCODER_FILES = \
    ["http://dl.fbaipublicfiles.com/BLINK/crossencoder_wiki_large.bin",
     "http://dl.fbaipublicfiles.com/BLINK/crossencoder_wiki_large.json"]

BLINK_FAISS_FLAT_INDEX = "http://dl.fbaipublicfiles.com/BLINK//faiss_flat_index.pkl"
BLINK_FAISS_HNSW_INDEX = "http://dl.fbaipublicfiles.com/BLINK/faiss_hnsw_index.pkl"

_config = {
    "test_entities": None,
    "test_mentions": None,
    "interactive": False,
    "top_k": 1,
    "biencoder_model": MODELS_CACHE_PATH + "biencoder_wiki_large.bin",
    "biencoder_config": MODELS_CACHE_PATH + "biencoder_wiki_large.json",
    "entity_catalogue": MODELS_CACHE_PATH + "entity.jsonl",
    "entity_encoding": MODELS_CACHE_PATH + "all_entities_large.t7",
    "crossencoder_model": MODELS_CACHE_PATH + "crossencoder_wiki_large.bin",
    "crossencoder_config": MODELS_CACHE_PATH + "crossencoder_wiki_large.json",
    "fast": True,
    "faiss_index": None,
    "index_path": None,
    "output_path": "logs/"
}


class BlinkIndex(str, Enum):
    FLAT = "flat"
    HNSW = "hnsw"
    NONE = None


blink_index2url = {BlinkIndex.FLAT: BLINK_FAISS_FLAT_INDEX, BlinkIndex.HNSW: BLINK_FAISS_HNSW_INDEX}


class LinkerBlink(Linker):
    """ Blink linker """

    def __init__(self, index=BlinkIndex.FLAT):
        """
        :param index: Index to use to perform the entity linking. One of:
            - BlinkIndex.FLAT
            - BlinkIndex.HNSW
        """
        super().__init__()

        if not pkgutil.find_loader("blink"):
            raise Exception("Blink module not installed. You need to install blink in order to use the Blink Linker."
                            "Install it with: pip install -e git+https://github.com/facebookresearch/BLINK.git#egg"
                            "=BLINK")
        self.index = index
        if index:
            _config['faiss_index'] = index.value
            _config['index_path'] = os.path.join(MODELS_CACHE_PATH, blink_index2url[index].split('/')[-1])
        self.config = argparse.Namespace(**_config)
        self.models = None
        self._wikipedia_id2local_id = None

    @property
    def entities_list(self) -> List[str]:
        """ Get list of entities """
        if len(self.models) < 6:
            raise Exception('model not yet initialized')
        return list(self.models[5].keys())

    @property
    def local_id2wikipedia_id(self):
        """ Get the Wikipedia ID from the label predicted """
        if len(self.models) < 9:
            raise Exception('model not yet initialized')
        if not self._wikipedia_id2local_id:
            self._wikipedia_id2local_id = {v: k for k, v in self.models[8].items()}
        return self._wikipedia_id2local_id

    def local_name2wikipedia_url(self, label: str) -> str:
        """ Get the Wikipedia URL of the label predicted to perform wikification

        :param label: Label to get URL of
        :return: URL of the Wikipedia article
        """
        wiki_id = self.local_id2wikipedia_id[self.models[5][label]]
        return f"https://en.wikipedia.org/wiki?curid={wiki_id}"

    def download_models(self):
        """ Download Blink files """
        files_to_download = BLINK_BI_ENCODER_FILES + [BLINK_ENTITIES]
        if self.index:
            files_to_download.append(blink_index2url[self.index])
        else:
            files_to_download.append(BLINK_ENTITIES_ENCODING)
        for f in files_to_download:
            download_file(f, output_dir=MODELS_CACHE_PATH)
        if not self.config.fast:
            for f in BLINK_CROSS_ENCODER_FILES:
                download_file(f, output_dir=MODELS_CACHE_PATH)

    def load_models(self):
        """ Load models """
        import blink.main_dense as main_dense
        if self.models is None:
            self.download_models()
            self.models = main_dense.load_models(self.config, logger=None)

    def predict(self, docs: Iterator[Doc], batch_size: Optional[Union[int, None]] = None) -> List[List[Span]]:
        """
        Perform the entity prediction
        :param docs: A list of spacy Document
        :param batch_size: The batch size
        :return: List Spans for each Document in docs
        """
        import blink.main_dense as main_dense
        self.load_models()
        data_to_link = []
        for doc_id, doc in enumerate(docs):
            for mention_id, mention in enumerate(doc._.mentions):
                data_to_link.append(
                    {
                        "id": doc_id,
                        "mention_id": mention_id,
                        "label": "unknown",
                        "label_id": -1,
                        "context_left": doc.text[:mention.start_char].lower(),
                        "mention": mention.text.lower(),
                        "context_right": doc.text[mention.end_char:].lower(),
                    })
        if not data_to_link:
            return []

        _, _, _, _, _, predictions, scores, = main_dense.run(self.config, None, *self.models, test_data=data_to_link)
        spans_annotations_values = {}
        for data, pred, score in zip(data_to_link, predictions, scores):
            doc = docs[data['id']]
            mention = doc._.mentions[data['mention_id']]
            span = Span(mention.start_char, mention.end_char, label=pred[0], score=score,
                        kb_id=self.local_name2wikipedia_url(pred[0]))
            if data['id'] in spans_annotations_values:
                spans_annotations_values['id'].append(span)
            else:
                spans_annotations_values['id'] = [span]

        spans_annotations_keys = sorted(spans_annotations_values.keys())
        spans_annotations = [spans_annotations_values[key] for key in spans_annotations_keys]

        return spans_annotations



import pkgutil
from typing import Optional, Iterator

from spacy.tokens.doc import Doc

from zshot.utils.data_models import Span
from zshot.mentions_extractor.mentions_extractor import MentionsExtractor
from zshot.mentions_extractor.utils import ExtractorType


class MentionsExtractorFlair(MentionsExtractor):
    """ Flair Mentions extractor """
    ALLOWED_CHUNKS = ("NP",)

    def __init__(self, extractor_type: Optional[ExtractorType] = ExtractorType.NER):
        """
        * Requires flair package to be installed *

        :param extractor_type: Type of extractor to get mentions. One of:
            - NER: to use Named Entity Recognition model to get the mentions
            - POS: to get the mentions based on the linguistics
        """
        if not pkgutil.find_loader("flair"):
            raise Exception("Flair module not installed. You need to install Flair for using this class."
                            "Install it with: pip install flair==0.11")

        super(MentionsExtractorFlair, self).__init__()

        self.extractor_type = extractor_type
        self.model = None

    def load_models(self):
        """ Load Flair model to perform the mentions extraction """
        if self.model is None:
            from flair.models import SequenceTagger
            if self.extractor_type == ExtractorType.NER:
                self.model = SequenceTagger.load("ner")
            else:
                self.model = SequenceTagger.load("chunk")

    def predict_pos_mentions(self, docs: Iterator[Doc], batch_size: Optional[int] = None):
        """ Predict mentions of docs using POS linguistics

        :param docs: Documents to get mentions of
        :param batch_size: Batch size to use
        :return: Spans of the mentions
        """
        from flair.data import Sentence
        sentences = [
            Sentence(str(doc), use_tokenizer=True) for doc in docs
        ]
        kwargs = {'mini_batch_size': batch_size} if batch_size else {}
        self.model.predict(sentences, **kwargs)

        spans = []
        for sent, doc in zip(sentences, docs):
            spans_tmp = []
            for i in range(len(sent.labels)):
                if sent.labels[i].value in self.ALLOWED_CHUNKS:
                    spans_tmp.append(Span(sent.labels[i].data_point.start_position,
                                          sent.labels[i].data_point.end_position))

            spans.append(spans_tmp)

        return spans

    def predict_ner_mentions(self, docs: Iterator[Doc], batch_size: Optional[int] = None):
        """ Predict mentions of docs using NER model

        :param docs: Documents to get mentions of
        :param batch_size: Batch size to use
        :return: Spans of the mentions
        """
        from flair.data import Sentence
        sentences = [
            Sentence(str(doc), use_tokenizer=True) for doc in docs
        ]
        kwargs = {'mini_batch_size': batch_size} if batch_size else {}
        self.model.predict(sentences, **kwargs)

        spans = []
        for sent, doc in zip(sentences, docs):
            sent_mentions = sent.get_spans('ner')
            spans_tmp = [
                Span(mention.start_position, mention.end_position, score=mention.score)
                for mention in sent_mentions
            ]
            spans.append(spans_tmp)

        return spans

    def predict(self, docs: Iterator[Doc], batch_size=None):
        """ Predict mentions in each document

        :param docs: Documents to get mentions of
        :param batch_size: Batch size to use
        :return: Spans of the mentions
        """
        self.load_models()
        if self.extractor_type == ExtractorType.NER:
            return self.predict_ner_mentions(docs, batch_size)
        else:
            return self.predict_pos_mentions(docs, batch_size)



from zshot.utils.data_models import Entity

ONTONOTES_ENTITIES = [Entity(name='NEG',
                             description="Coal, water, oil, etc. are normally used for traditional electricity "
                                         "generation. However using liquefied natural gas as fuel for joint "
                                         "circulatory electircity generation has advantages. The chief financial "
                                         "officer is the only one there taking the fall. It has a very talented team, "
                                         "eh. What will happen to the wildlife? I just tell them, you've got to change."
                                         " They're here to stay. They have no insurance on their cars. "
                                         "What else would you like? Whether holding an international cultural event "
                                         "or setting the city's cultural policies, she always asks for the "
                                         "participation or input of other cities and counties.",
                             vocabulary=None),
                      Entity(name='PERSON',
                             description='These are proper names of people, including fictional people, first names, '
                                         'last names, individual or family names, unique nicknames. '
                                         'Generational markers (Jr., IV) are included in the extent, while honorifics '
                                         '(Ms., Dr.) and occupational titles (President, Secretary) are NOT included.',
                             vocabulary=None),
                      Entity(name='NORP',
                             description='This type represents adjectival forms of GPE and Location names, as well '
                                         'as adjectival forms of named religions, heritage and political affiliation. '
                                         'Also marked are head words which refer to people using the name of an entity '
                                         'with which they are affiliated, often a GPE or Organization. The distinction '
                                         'between NORP and other types is morphological. "American" and "Americans" '
                                         'are adjectival nationalities, while "America" and "US" are GPEs, '
                                         'regardless of context.',
                             vocabulary=None),
                      Entity(name='FAC',
                             description='Names of man-made structures: infrastructure (streets, bridges), buildings, '
                                         'monuments, etc. belong to this type. Buildings that are referred to using '
                                         'the name of the company or organization that uses them should be marked as '
                                         'FAC when they refer to the physical structure of the building itself, '
                                         'usually in a locative way: "I\'m reporting live from right outside '
                                         '[Massachusetts General Hospital]"',
                             vocabulary=None),
                      Entity(name='ORG',
                             description='These are names of companies, government agencies, educational institutions, '
                                         'sport teams. Names of hospitals, museums, and libraries should be marked, '
                                         'unless the mentions are clearly referring to the building in a locative way.'
                                         ' Adjectival forms of organization names are included, as are metonymic '
                                         'mentions of the buildings or locations associated with these organizations.'
                                         ' A group, team, force, etc. must be officially sanctioned in some way to be '
                                         'classified as an Organization. Organized crime groups, such as the Mafia, '
                                         'are not marked. Terrorist groups such as Al-Qaeda,'
                                         ' however, should be marked.',
                             vocabulary=None),
                      Entity(name='GPE',
                             description='Names of countries, cities, states, provinces, municipalities. '
                                         'In cases where the GPE name is modified, such as "southern California," '
                                         '[California] is marked as a GPE name and there is NO other markup.',
                             vocabulary=None),
                      Entity(name='LOC',
                             description='Names of geographical locations other than GPEs. These include '
                                         'mountain ranges, coasts, borders, planets, geo-coordinates, '
                                         'bodies of water. Also included in this category are named regions such as '
                                         'the Middle East, areas, neighborhoods, continents and regions of continents.'
                                         ' Do NOT mark deictics or other non-proper nouns: here, there, everywhere,'
                                         ' etc. As with GPEs, directional modifiers such as "southern" are only '
                                         'marked when they are part of the location name itself.',
                             vocabulary=None),
                      Entity(name='PRODUCT',
                             description='This can be name of any product, generally a model name or model '
                                         'name and number. Named foods are also included. Credit cards, checking'
                                         ' accounts, CDs, and credit plans are NOT marked. References that include'
                                         ' manufacturer and product should be marked as two separate named entities,'
                                         ' ORG + PRODUCT: [Apple] [iPod], [Dell] [Inspiron], [Ford] [Mustang].',
                             vocabulary=None),
                      Entity(name='DATE',
                             description='Used to classify a reference to a date or period, etc. Age also falls '
                                         'under this category, even when it’s a noun phrase referring to a person: '
                                         'the 5-year-old, 5 years old, Jane Doe, 5, etc. Extent should include '
                                         'modifiers & prepositions that denote specific time, such as [2 days ago], '
                                         '[the past two days], but not those that mark duration, such as "for '
                                         '[2 days]." Do not separate mentions into their component parts: [November'
                                         ' 2, 2001] and [from the fall of 1962 through the spring of 1967] should '
                                         'be marked in their entirety, without separate markups for "November 2," '
                                         '"2001, "the fall," "1962," "the spring," and "1967." Dates that are part'
                                         ' of rate expressions such as "once per day" or "twice a year" '
                                         'should NOT be marked.',
                             vocabulary=None),
                      Entity(name='TIME',
                             description='Any time ending with "a.m." or "p.m." If the "a.m." or "p.m." is explicit, '
                                         'it must be tagged along with the numbers. Other times of day (units '
                                         'smaller than a day) and time durations less than 24 hours are also '
                                         'marked: morning, noon, night, 3 hours, a minute. Do not separate mentions'
                                         ' into their component parts: [the evening of July 4 th ] and [5:00am, '
                                         'April 5, 2008] should be marked in their entirety, without separate markups'
                                         ' for "evening," "July 4 th ," "5:00am," and "April 5, 2008"',
                             vocabulary=None),
                      Entity(name='PERCENT',
                             description='Any percentage. A percent symbol or the actual word percent must be '
                                         'explicit and included in the extent. If the percent is implicit, the '
                                         'number should be marked CARDINAL.',
                             vocabulary=None),
                      Entity(name='MONEY',
                             description='Any monetary value including all monetary denominations. The monetary unit'
                                         ' must be explicit and included in the extent. If the monetary unit is '
                                         'implicit, the number should be marked CARDINAL. Only values should be '
                                         'tagged—generic references to money should not. For example, in "money '
                                         'invested overseas," there is no markup for "money." In rate expressions'
                                         ' such as "$ per unit," the unit should not be included in the extent. '
                                         'For example, in "$3 per share," the extent is [$3].',
                             vocabulary=None),
                      Entity(name='QUANTITY',
                             description='Used to classify measurements with standardized units. If the unit of'
                                         ' measurement is implicit or non-standard (3 handfuls, 2 football fields,'
                                         ' 10 points), the number should be marked CARDINAL. One exception to this'
                                         ' rule is formulaic references to the age, height, and weight of a person:'
                                         ' Joe Smith, 44, five ten, two twenty. In this instance, [five ten] and'
                                         ' [two twenty] should be marked QUANTITY. (While [44] should be marked DATE)',
                             vocabulary=None),
                      Entity(name='ORDINAL', description='All ordinal numbers, including adverbials.', vocabulary=None),
                      Entity(name='CARDINAL',
                             description='Numerals, including whole numbers, fractions, and decimals, that provide'
                                         ' a count or quantity and do not fall under a unit of measurement, money,'
                                         ' percent, date or time. For "Nasdaq composite fill [1.39] to [451.37]." '
                                         'the numbers are marked CARDINAL because there is no monetary unit. '
                                         'Headless numerical phrases are also covered in this category: "reducing '
                                         'employment from [18,000] to [16,000]." Numbers identifying list items '
                                         'should also be included. Pronominal mentions of "one" should not be tagged.',
                             vocabulary=None),
                      Entity(name='EVENT',
                             description='Named hurricanes, battles, wars, sports events, attacks. Metonymic mentions '
                                         '(marked with a ~) of the date or location of an event, '
                                         'or of the organization(s) involved, are included.',
                             vocabulary=None),
                      Entity(name='WORK_OF_ART',
                             description='Titles of books, songs, television programs and other creations. '
                                         'Also includes awards. These are usually surrounded by quotation '
                                         'marks in the article (though the quotations are not included in '
                                         'the annotation). Newspaper headlines should only be marked if they '
                                         'are referential. In other words the headline of the article being '
                                         'annotated should not be marked but if in the body of the text here is '
                                         'a reference to an article, then it is markable as a work of art.',
                             vocabulary=None),
                      Entity(name='LAW',
                             description='Any document that has been made into a law, including named '
                                         'treaties and sections and chapters of named legal documents.',
                             vocabulary=None),
                      Entity(name='LANGUAGE', description='Any named language.', vocabulary=None)]



import zlib
from typing import Optional, List
from pydantic import BaseModel


class Entity(BaseModel):
    name: str
    description: str
    vocabulary: Optional[List[str]] = None

    def __hash__(self):
        self_repr = f"{self.__class__.__name__}.{str(self.__dict__)}"
        return zlib.crc32(self_repr.encode())



import os
import pickle as pkl
import zlib
from abc import ABC, abstractmethod
from typing import Iterator, List, Optional, Union

from spacy.tokens import Doc
from spacy.util import ensure_path

from zshot.utils.data_models import Entity, Span
from zshot.utils.alignment_utils import filter_overlapping_spans, spacy_token_offsets


class Linker(ABC):
    """
    Linker define a standard interface for entity linking. A Linker may relay on existing
    extracted mentions or perform end-2-end extraction
    """

    def __init__(self):
        self._entities = None
        self._is_end2end = False

    def set_kg(self, entities: Iterator[Entity]):
        """
        Set entities that linker can use
        :param entities: The list of entities
        """
        self._entities = entities

    @property
    def entities(self) -> List[Entity]:
        """ Entities to link to """
        return self._entities

    def load_models(self):
        """
        Load the model
        :return:
        """
        pass

    @abstractmethod
    def predict(self, docs: Iterator[Doc], batch_size: Optional[Union[int, None]] = None) -> List[List[Span]]:
        """
        Perform the entity prediction
        :param docs: A list of spacy Document
        :param batch_size: The batch size
        :return: List Spans for each Document in docs
        """
        pass

    @property
    def is_end2end(self) -> bool:
        return self._is_end2end

    @is_end2end.setter
    def is_end2end(self, value):
        self._is_end2end = value

    @staticmethod
    def version() -> str:
        return "v1"

    def link(self, docs: Iterator[Doc], batch_size: Optional[Union[int, None]] = None):
        """
        Perform the entity linking. Call the predict function and add entities to the Spacy Docs
        :param docs: A list of spacy Document
        :param batch_size: The batch size
        :return:
        """
        predictions_spans = self.predict(docs, batch_size)

        for d, preds in zip(docs, predictions_spans):
            d._.spans = preds
            d.ents = map(lambda p: p.to_spacy_span(d), filter_overlapping_spans(preds, list(d),
                                                                                tokens_offsets=spacy_token_offsets(d)))
            # d.spans = map(lambda p: p.to_spacy_span(d), preds)

    @staticmethod
    def _get_serialize_file(path):
        return os.path.join(path, "linker.pkl")

    @staticmethod
    def _get_config_file(path):
        path = os.path.join(path, "linker.json")
        path = ensure_path(path)
        return path

    @classmethod
    def from_disk(cls, path, exclude=()):
        serialize_file = cls._get_serialize_file(path)
        with open(serialize_file, "rb") as f:
            return pkl.load(f)

    def to_disk(self, path):
        serialize_file = self._get_serialize_file(path)
        with open(serialize_file, "wb") as f:
            return pkl.dump(self, f)

    def __hash__(self):
        self_repr = f"{self.__class__.__name__}.{self.version()}.{str(self.__dict__)}"
        return zlib.crc32(self_repr.encode())



from typing import Optional, List, Union

import spacy
from evaluate import EvaluationModule
from prettytable import PrettyTable

from zshot.evaluation import load_medmentions, load_ontonotes
from zshot.evaluation.dataset.dataset import DatasetWithEntities
from zshot.evaluation.evaluator import ZeroShotTokenClassificationEvaluator, MentionsExtractorEvaluator
from zshot.evaluation.pipeline import LinkerPipeline, MentionsExtractorPipeline


def evaluate(nlp: spacy.language.Language,
             datasets: Union[DatasetWithEntities, List[DatasetWithEntities]],
             splits: Optional[Union[str, List[str]]] = None,
             metric: Optional[Union[str, EvaluationModule]] = None,
             batch_size: Optional[int] = 16) -> str:
    """ Evaluate a spacy zshot model

    :param nlp: Spacy Language pipeline with ZShot components
    :param datasets: Dataset or list of datasets to evaluate
    :param splits: Optional. Split or list of splits to evaluate. All splits available by default
    :param metric: Metrics to use in evaluation.
        Options available: precision, recall, f1-score-micro, f1-score-macro. All by default
    :return: Result of the evaluation. Dict with precision, recall and f1-score for each component
    :param batch_size: the batch size
    """
    linker_evaluator = ZeroShotTokenClassificationEvaluator("token-classification")
    mentions_extractor_evaluator = MentionsExtractorEvaluator("token-classification")

    if type(splits) == str:
        splits = [splits]

    result = {}
    field_names = ["Metric"]
    for dataset_name in datasets:
        if dataset_name.lower() == "medmentions":
            dataset = load_medmentions()
        else:
            dataset = load_ontonotes()

        for split in splits:
            field_name = f"{dataset_name} {split}"
            field_names.append(field_name)
            nlp.get_pipe("zshot").entities = dataset[split].entities
            if nlp.get_pipe("zshot").linker:
                pipe = LinkerPipeline(nlp, batch_size)
                result.update(
                    {
                        field_name: {
                            'linker': linker_evaluator.compute(pipe, dataset[split], metric=metric)
                        }
                    }
                )
            # TODO: Add support for mentions_extractor pipelines and evaluation
            if nlp.get_pipe("zshot").mentions_extractor:
                pipe = MentionsExtractorPipeline(nlp, batch_size)
                result.update(
                    {
                        field_name: {
                            'mentions_extractor': mentions_extractor_evaluator.compute(pipe, dataset[split],
                                                                                       metric=metric)
                        }
                    }
                )

    table = PrettyTable()
    table.field_names = field_names
    rows = []
    if nlp.get_pipe("zshot").linker:
        linker_precisions = []
        linker_recalls = []
        linker_micros = []
        linker_macros = []
        linker_accuracies = []
        for field_name in field_names:
            if field_name == "Metric":
                continue
            linker_precisions.append("{:.2f}%".format(result[field_name]['linker']['overall_precision_macro'] * 100))
            linker_recalls.append("{:.2f}%".format(result[field_name]['linker']['overall_recall_macro'] * 100))
            linker_accuracies.append("{:.2f}%".format(result[field_name]['linker']['overall_accuracy'] * 100))
            linker_micros.append("{:.2f}%".format(result[field_name]['linker']['overall_f1_micro'] * 100))
            linker_macros.append("{:.2f}%".format(result[field_name]['linker']['overall_f1_macro'] * 100))

        rows.append(["Linker Precision"] + linker_precisions)
        rows.append(["Linker Recall"] + linker_recalls)
        rows.append(["Linker Accuracy"] + linker_accuracies)
        rows.append(["Linker F1-score micro"] + linker_micros)
        rows.append(["Linker F1-score macro"] + linker_macros)

    if nlp.get_pipe("zshot").mentions_extractor:
        mentions_extractor_precisions = []
        mentions_extractor_recalls = []
        mentions_extractor_micros = []
        mentions_extractor_accuracies = []
        mentions_extractor_macros = []
        for field_name in field_names:
            if field_name == "Metric":
                continue
            mentions_extractor_precisions.append(
                "{:.2f}%".format(result[field_name]['mentions_extractor']['overall_precision_macro'] * 100))
            mentions_extractor_recalls.append(
                "{:.2f}%".format(result[field_name]['mentions_extractor']['overall_recall_macro'] * 100))
            mentions_extractor_accuracies.append(
                "{:.2f}%".format(result[field_name]['mentions_extractor']['overall_accuracy'] * 100))
            mentions_extractor_micros.append(
                "{:.2f}%".format(result[field_name]['mentions_extractor']['overall_f1_micro'] * 100))
            mentions_extractor_macros.append(
                "{:.2f}%".format(result[field_name]['mentions_extractor']['overall_f1_macro'] * 100))

        rows.append(["Mentions extractor Precision"] + mentions_extractor_precisions)
        rows.append(["Mentions extractor Recall"] + mentions_extractor_recalls)
        rows.append(["Mentions extractor Accuracy"] + mentions_extractor_accuracies)
        rows.append(["Mentions extractor F1-score micro"] + mentions_extractor_micros)
        rows.append(["Mentions extractor F1-score macro"] + mentions_extractor_macros)

    table.add_rows(rows)

    return table.get_string()



import pkgutil

import spacy

from zshot.mentions_extractor import MentionsExtractorFlair
from zshot.mentions_extractor.mentions_extractor_flair import ExtractorType
from zshot.tests.config import EX_DOCS
from zshot import PipelineConfig


def test_flair_ner_mentions_extractor():
    if not pkgutil.find_loader("flair"):
        return
    nlp = spacy.blank("en")
    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorFlair(ExtractorType.NER))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])
    assert doc.ents == ()
    assert len(doc._.mentions) > 0


def test_custom_flair_mentions_extractor():
    nlp = spacy.blank("en")
    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorFlair(ExtractorType.NER))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])
    assert doc.ents == ()
    assert len(doc._.mentions) > 0


def test_flair_pos_mentions_extractor():
    if not pkgutil.find_loader("flair"):
        return
    nlp = spacy.blank("en")

    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorFlair(ExtractorType.POS))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])
    assert doc.ents == ()
    assert len(doc._.mentions) > 0


def test_flair_ner_mentions_extractor_pipeline():
    if not pkgutil.find_loader("flair"):
        return
    nlp = spacy.blank("en")
    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorFlair(ExtractorType.NER))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    docs = [doc for doc in nlp.pipe(EX_DOCS)]
    assert all(doc.ents == () for doc in docs)
    assert all(len(doc._.mentions) > 0 for doc in docs)


def test_flair_pos_mentions_extractor_pipeline():
    if not pkgutil.find_loader("flair"):
        return
    nlp = spacy.blank("en")
    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorFlair(ExtractorType.POS))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    docs = [doc for doc in nlp.pipe(EX_DOCS)]
    assert all(doc.ents == () for doc in docs)
    assert all(len(doc._.mentions) > 0 for doc in docs)



class LinkerPipeline:
    def __init__(self, nlp, batch_size=100):
        self.nlp = nlp
        self.task = 'token-classification'
        self.batch_size = batch_size

    def __call__(self, *args, **kwargs):
        res = []
        docs = self.nlp.pipe(args[0], batch_size=self.batch_size)
        for doc in docs:
            res_doc = []
            for span in doc._.spans:
                label = {
                    'entity': span.label, 'score': span.score,
                    'word': doc.text[span.start:span.end],
                    'start': span.start, 'end': span.end
                }
                res_doc.append(label)
            res.append(res_doc)

        return res


class MentionsExtractorPipeline:
    def __init__(self, nlp, batch_size=100):
        self.nlp = nlp
        self.task = 'token-classification'
        self.batch_size = batch_size

    def __call__(self, *args, **kwargs):
        res = []
        docs = self.nlp.pipe(args[0], batch_size=self.batch_size)
        for doc in docs:
            res_doc = []
            for span in doc._.mentions:
                label = {
                    'entity': "MENTION",
                    'word': doc.text[span.start_char:span.end_char],
                    'start': span.start_char, 'end': span.end_char
                }
                res_doc.append(label)
            res.append(res_doc)

        return res



from typing import Iterator

import spacy
from spacy.tokens.doc import Doc

from zshot import MentionsExtractor, PipelineConfig
from zshot.utils.data_models import Span
from zshot.tests.config import EX_DOCS, EX_ENTITIES


class DummyMentionsExtractor(MentionsExtractor):
    def predict(self, docs: Iterator[Doc], batch_size=None):
        return [
            [Span(0, len(doc.text) - 1)]
            for doc in docs
        ]


class DummyMentionsExtractorWithNER(MentionsExtractor):
    @property
    def require_existing_ner(self) -> bool:
        return True

    def predict(self, docs: Iterator[Doc], batch_size=None):
        return [
            [Span(0, len(doc.text) - 1)]
            for doc in docs
        ]


class DummyMentionsExtractorWithEntities(MentionsExtractor):
    def predict(self, docs: Iterator[Doc], batch_size=None):
        return [
            [Span(0, len(doc.text) - 1)]
            for idx, doc in enumerate(docs)
        ]


def test_dummy_mentions_extractor():
    nlp = spacy.blank("en")
    config_zshot = PipelineConfig(mentions_extractor=DummyMentionsExtractor())
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])
    assert doc.ents == ()
    assert len(doc._.mentions) > 0


def test_dummy_mentions_extractor_with_entities_config():
    nlp = spacy.blank("en")
    config_zshot = PipelineConfig(mentions_extractor=DummyMentionsExtractorWithEntities(),
                                  entities=EX_ENTITIES)
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])
    assert doc.ents == ()
    assert len(doc._.mentions) > 0



from typing import List, Iterator, Tuple

import spacy
from datasets import Dataset
from spacy.tokens import Doc

from zshot import PipelineConfig, Linker
from zshot.evaluation.evaluator import ZeroShotTokenClassificationEvaluator
from zshot.evaluation.pipeline import LinkerPipeline
from zshot.utils.alignment_utils import AlignmentMode
from zshot.utils.data_models import Entity, Span

ENTITIES = [
    Entity(name="FAC", description="A facility"),
    Entity(name="LOC", description="A location"),
]


class DummyLinker(Linker):

    def __init__(self, predictions: List[Tuple[str, str, float]]):
        super().__init__()
        self.predictions = predictions

    def predict(self, docs: Iterator[Doc], batch_size=None):
        sentences = []
        for doc in docs:
            preds = []
            for span, label, score in self.predictions:
                if span in doc.text:
                    preds.append(Span(doc.text.find(span), doc.text.find(span) + len(span), label=label, score=score))
            sentences.append(preds)

        return sentences


def get_pipe(predictions: List[Tuple[str, str, float]]):
    nlp = spacy.blank("en")
    nlp_config = PipelineConfig(
        linker=DummyLinker(predictions),
        entities=ENTITIES
    )

    nlp.add_pipe("zshot", config=nlp_config, last=True)

    return LinkerPipeline(nlp)


def get_spans_predictions(span: str, label: str, sentence: str):
    return [{'start': sentence.find(span),
             'end': sentence.find(span) + len(span),
             'entity': label,
             'score': 1}]


def get_dataset(gt: List[List[str]], sentence: List[str]):
    data_dict = {
        'tokens': [s.split(" ") for s in sentence],
        'ner_tags': gt,
    }
    dataset = Dataset.from_dict(data_dict)
    dataset.entities = ENTITIES
    return dataset


class TestZeroShotTokenClassificationEvaluation:

    def test_preprocess(self):
        sentences = ['New York is beautiful']
        gt = [['B-FAC', 'I-FAC', 'O', 'O']]

        predictions = [get_spans_predictions('New York', 'FAC', sentences[0])]
        custom_evaluator = ZeroShotTokenClassificationEvaluator("token-classification")

        preds = custom_evaluator.predictions_processor(predictions, [['New', 'York', 'is', 'beautiful']], join_by=" ")
        assert preds['predictions'] == gt

    def test_prediction_token_based_evaluation_all_matching(self):
        sentences = ['New York is beautiful']
        gt = [['B-FAC', 'I-FAC', 'O', 'O']]

        dataset = get_dataset(gt, sentences)

        custom_evaluator = ZeroShotTokenClassificationEvaluator("token-classification")
        metrics = custom_evaluator.compute(get_pipe([('New York', 'FAC', 1)]), dataset, "seqeval")

        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_f1"]) == 1.0
        assert float(metrics["overall_accuracy"]) == 1.0

    def test_prediction_token_based_evaluation_overlapping_spans(self):
        sentences = ['New York is beautiful']
        gt = [['B-FAC', 'I-FAC', 'O', 'O']]

        dataset = get_dataset(gt, sentences)

        custom_evaluator = ZeroShotTokenClassificationEvaluator("token-classification")
        metrics = custom_evaluator.compute(get_pipe([('New York', 'FAC', 1), ('York', 'LOC', 0.7)]), dataset, "seqeval")

        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_f1"]) == 1.0
        assert float(metrics["overall_accuracy"]) == 1.0

    def test_prediction_token_based_evaluation_partial_match_spans_expand(self):
        sentences = ['New York is beautiful']
        gt = [['B-FAC', 'I-FAC', 'O', 'O']]

        dataset = get_dataset(gt, sentences)

        custom_evaluator = ZeroShotTokenClassificationEvaluator("token-classification",
                                                                alignment_mode=AlignmentMode.expand)
        pipe = get_pipe([('New Yo', 'FAC', 1)])
        metrics = custom_evaluator.compute(pipe, dataset, "seqeval")

        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_f1"]) == 1.0
        assert float(metrics["overall_accuracy"]) == 1.0

    def test_prediction_token_based_evaluation_partial_match_spans_contract(self):
        sentences = ['New York is beautiful']
        gt = [['B-FAC', 'I-FAC', 'O', 'O']]

        dataset = get_dataset(gt, sentences)

        custom_evaluator = ZeroShotTokenClassificationEvaluator("token-classification",
                                                                alignment_mode=AlignmentMode.contract)
        pipe = get_pipe([('New York i', 'FAC', 1)])
        metrics = custom_evaluator.compute(pipe, dataset, "seqeval")

        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_f1"]) == 1.0
        assert float(metrics["overall_accuracy"]) == 1.0

    def test_prediction_token_based_evaluation_partial_and_overlapping_spans(self):
        sentences = ['New York is beautiful']
        gt = [['B-FAC', 'I-FAC', 'O', 'O']]

        dataset = get_dataset(gt, sentences)

        custom_evaluator = ZeroShotTokenClassificationEvaluator("token-classification",
                                                                alignment_mode=AlignmentMode.contract)
        pipe = get_pipe([('New York i', 'FAC', 1), ('w York', 'LOC', 0.7)])
        metrics = custom_evaluator.compute(pipe, dataset, "seqeval")

        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_precision"]) == 1.0
        assert float(metrics["overall_f1"]) == 1.0
        assert float(metrics["overall_accuracy"]) == 1.0



from typing import Optional, Iterator

from spacy.tokens.doc import Doc

from zshot.utils.data_models import Span
from zshot.mentions_extractor.mentions_extractor import MentionsExtractor
from zshot.mentions_extractor.utils import ExtractorType


class MentionsExtractorSpacy(MentionsExtractor):
    """ SpaCy mentions extractor """
    ALLOWED_POS = ("NOUN", "PROPN")
    ALLOWED_DEP = ("compound", "pobj", "dobj", "nsubj", "attr", "appos")
    COMPOUND_DEP = "compound"

    EXCLUDE_NER = ("CARDINAL", "DATE", "ORDINAL", "PERCENT", "QUANTITY", "TIME")

    def __init__(self, extractor_type: Optional[ExtractorType] = ExtractorType.NER):
        """
        :param extractor_type: Type of extractor to get mentions. One of:
            - NER: to use Named Entity Recognition model to get the mentions
            - POS: to get the mentions based on the linguistics
        """
        super(MentionsExtractorSpacy, self).__init__()
        self.extractor_type = extractor_type

    @property
    def require_existing_ner(self) -> bool:
        """ If the type of the extractor is NER the existing NER is required """
        return self.extractor_type == ExtractorType.NER

    def predict_pos_mentions(self, docs: Iterator[Doc], batch_size: Optional[int] = None):
        """ Predict mentions of docs using POS linguistics

        :param docs: Documents to get mentions of
        :param batch_size: Batch size to use
        :return: Spans of the mentions
        """
        spans = []
        for doc in docs:
            skip = -1
            spans_tmp = []
            for i, tok in enumerate(doc):
                if 0 < i < skip:
                    continue

                if tok.pos_ in self.ALLOWED_POS and tok.dep_ in self.ALLOWED_DEP:
                    if tok.dep_ == self.COMPOUND_DEP:
                        spans_tmp.append(Span(tok.idx, tok.head.idx + len(tok.head)))
                        skip = tok.head.i + 1
                    else:
                        spans_tmp.append(Span(tok.idx, tok.idx + len(tok)))
            spans.append(spans_tmp)

        return spans

    def predict_ner_mentions(self, docs: Iterator[Doc], batch_size: Optional[int] = None):
        """ Predict mentions of docs using NER model

        :param docs: Documents to get mentions of
        :param batch_size: Batch size to use
        :return: Spans of the mentions
        """
        spans = [
            [
                Span(ent.start_char, ent.end_char)
                for ent in doc.ents if ent.label_ not in self.EXCLUDE_NER
            ]
            for doc in docs
        ]
        for doc in docs:
            doc.ents = []

        return spans

    def predict(self, docs: Iterator[Doc], batch_size=None):
        """ Predict mentions of docs

        :param docs: Documents to get mentions of
        :param batch_size: Batch size to use
        :return: Spans of the mentions
        """
        if self.extractor_type == ExtractorType.NER:
            return self.predict_ner_mentions(docs, batch_size)
        else:
            return self.predict_pos_mentions(docs, batch_size)



from zshot.mentions_extractor.mentions_extractor_flair import MentionsExtractorFlair  # noqa: F401
from zshot.mentions_extractor.mentions_extractor_spacy import MentionsExtractorSpacy  # noqa: F401
from zshot.mentions_extractor.mentions_extractor import MentionsExtractor  # noqa: F401



import spacy

from zshot import PipelineConfig
from zshot.linker.linker_regen.linker_regen import LinkerRegen
from zshot.mentions_extractor import MentionsExtractorSpacy
from zshot.tests.config import EX_DOCS, EX_ENTITIES


def test_regen_linker():
    nlp = spacy.load("en_core_web_sm")
    config = PipelineConfig(
        mentions_extractor=MentionsExtractorSpacy(),
        linker=LinkerRegen(),
        entities=EX_ENTITIES
    )
    nlp.add_pipe("zshot", config=config, last=True)
    assert "zshot" in nlp.pipe_names

    doc = nlp(EX_DOCS[1])
    assert len(doc.ents) > 0


def test_regen_linker_pipeline():
    nlp = spacy.load("en_core_web_sm")
    config = PipelineConfig(
        mentions_extractor=MentionsExtractorSpacy(),
        linker=LinkerRegen(),
        entities=EX_ENTITIES
    )
    nlp.add_pipe("zshot", config=config, last=True)
    assert "zshot" in nlp.pipe_names

    docs = [doc for doc in nlp.pipe(EX_DOCS)]
    assert all(len(doc.ents) > 0 for doc in docs)



import os
import pathlib

from appdata import AppDataPaths

MODELS_CACHE_PATH = os.getenv("MODELS_CACHE_PATH") if "MODELS_CACHE_PATH" in os.environ \
    else AppDataPaths(f"{pathlib.Path(__file__).stem}").app_data_path + "/"



import spacy

from zshot import PipelineConfig, displacy
from zshot.linker import LinkerRegen
from zshot.mentions_extractor import MentionsExtractorSpacy
from zshot.utils.data_models import Entity

nlp = spacy.load("en_core_web_sm")

nlp_config = PipelineConfig(
    mentions_extractor=MentionsExtractorSpacy(),
    linker=LinkerRegen(),
    entities=[
        Entity(name="Paris",
               description="Paris is located in northern central France, in a north-bending arc of the river Seine"),
        Entity(name="IBM",
               description="International Business Machines Corporation (IBM) is an American multinational technology corporation headquartered in Armonk, New York"),
        Entity(name="New York", description="New York is a city in U.S. state"),
        Entity(name="Florida", description="southeasternmost U.S. state"),
        Entity(name="American",
               description="American, something of, from, or related to the United States of America, commonly known as the United States or America"),
        Entity(name="Chemical formula",
               description="In chemistry, a chemical formula is a way of presenting information about the chemical proportions of atoms that constitute a particular chemical compound or molecule"),
        Entity(name="Acetamide",
               description="Acetamide (systematic name: ethanamide) is an organic compound with the formula CH3CONH2. It is the simplest amide derived from acetic acid. It finds some use as a plasticizer and as an industrial solvent."),
        Entity(name="Armonk",
               description="Armonk is a hamlet and census-designated place (CDP) in the town of North Castle, located in Westchester County, New York, United States."),
        Entity(name="Acetic Acid",
               description="Acetic acid, systematically named ethanoic acid, is an acidic, colourless liquid and organic compound with the chemical formula CH3COOH"),
        Entity(name="Industrial solvent",
               description="Acetamide (systematic name: ethanamide) is an organic compound with the formula CH3CONH2. It is the simplest amide derived from acetic acid. It finds some use as a plasticizer and as an industrial solvent."),
    ]
)

nlp.add_pipe("zshot", config=nlp_config, last=True)

text = "International Business Machines Corporation (IBM) is an American multinational technology corporation" \
       " headquartered in Armonk, New York, with operations in over 171 countries."

doc = nlp()
displacy.serve(doc, style="ent", port=8888)



from datasets import ClassLabel, load_dataset, DatasetDict

from zshot.evaluation.dataset.dataset import DatasetWithEntities
from zshot.evaluation.dataset.ontonotes.entities import ONTONOTES_ENTITIES

LABELS = ONTONOTES_ENTITIES
labels = ClassLabel(num_classes=37,
                    names=["O", "B-PERSON", "I-PERSON", "B-NORP", "I-NORP", "B-FAC", "I-FAC",
                           "B-ORG", "I-ORG", "B-GPE", "I-GPE", "B-LOC", "I-LOC",
                           "B-PRODUCT", "I-PRODUCT", "B-DATE", "I-DATE", "B-TIME", "I-TIME",
                           "B-PERCENT", "I-PERCENT", "B-MONEY", "I-MONEY", "B-QUANTITY", "I-QUANTITY",
                           "B-ORDINAL", "I-ORDINAL", "B-CARDINAL", "I-CARDINAL", "B-EVENT", "I-EVENT",
                           "B-WORK_OF_ART", "I-WORK_OF_ART", "B-LAW", "I-LAW", "B-LANGUAGE", "I-LANGUAGE"])

CLASSES_PER_SPLIT = {
    "train": ["PERSON", "GPE", "ORG", "DATE"],
    "validation": ["NORP", "MONEY", 'ORDINAL', "PERCENT", "EVENT", "PRODUCT", "LAW"],
    "test": ["CARDINAL", "TIME", "LOC", "WORK_OF_ART", "FAC", "QUANTITY", "LANGUAGE"]
}
TRIVIAL_CLASSES = ["ORDINAL", "QUANTITY", "MONEY", "PERCENT", "CARDINAL", "LANGUAGE", "TIME"]


def remove_other_tasks(sentence):
    if 'pos_tags' in sentence:
        del sentence['pos_tags']
    if 'parse_tree' in sentence:
        del sentence['parse_tree']
    if 'predicate_framenet_ids' in sentence:
        del sentence['predicate_framenet_ids']
    if 'word_senses' in sentence:
        del sentence['word_senses']
    if 'speaker' in sentence:
        del sentence['speaker']
    if 'predicate_lemmas' in sentence:
        del sentence['predicate_lemmas']
    if 'coref_spans' in sentence:
        del sentence['coref_spans']
    if 'srl_frames' in sentence:
        del sentence['srl_frames']
    return sentence


def is_not_empty(sentence):
    return not all([s == 0 for s in sentence['named_entities']])


def remove_out_of_split(sentence, split):
    for i, ent in enumerate(sentence['named_entities']):
        label = labels.int2str(ent)
        if label == 'O' or label[2:] in TRIVIAL_CLASSES or label[2:] not in CLASSES_PER_SPLIT[split]:
            sentence['named_entities'][i] = 0
    return sentence


def load_ontonotes() -> DatasetDict[DatasetWithEntities]:
    dataset_zs = load_dataset("conll2012_ontonotesv5", "english_v12")
    ontonotes_zs = DatasetDict()

    for split in dataset_zs:
        dataset_zs[split] = dataset_zs[split].map(lambda example, idx: {
            "sentences": [remove_out_of_split(s, split) for s in example['sentences']]
        }, with_indices=True)
        dataset_zs[split] = dataset_zs[split].map(lambda example, idx: {
            "sentences": list(filter(is_not_empty, example['sentences']))
        }, with_indices=True)

        tokens = []
        ner_tags = []
        for example in dataset_zs[split]:
            tokens += [s['words'] for s in example['sentences']]
            ner_tags += [[labels.int2str(ent) for ent in s['named_entities']] for s in example['sentences']]

        split_entities = [ent for ent in ONTONOTES_ENTITIES
                          if ent.name in ['NEG'] + CLASSES_PER_SPLIT[split] and ent.name not in TRIVIAL_CLASSES]

        ontonotes_zs[split] = DatasetWithEntities.from_dict({
            'tokens': tokens,
            'ner_tags': ner_tags
        }, split=split, entities=split_entities)

    return ontonotes_zs



import os
import pickle as pkl
import warnings

import zlib
from abc import ABC, abstractmethod
from spacy.tokens import Doc
from typing import List, Iterator

from spacy.util import ensure_path

from zshot.utils.data_models import Entity
from zshot.utils.data_models import Span


class MentionsExtractor(ABC):

    def __init__(self):
        self._entities = None

    def set_kg(self, entities: Iterator[Entity]):
        """
        Set entities that mention extractor can use
        :param entities: The list of entities
        """
        self._entities = entities

    @property
    def entities(self) -> List[Entity]:
        return self._entities

    def load_models(self):
        """
        Load the model
        :return:
        """
        pass

    def extract_mentions(self, docs: Iterator[Doc], batch_size=None):
        """
        Perform the mentions extraction. Call the predict function and add the mentions to the Spacy Doc
        :param docs: A list of spacy Document
        :param batch_size: The batch size
        :return:
        """
        predictions_spans = self.predict(docs, batch_size)
        for doc, doc_preds in zip(docs, predictions_spans):
            doc_pred_spans = [
                doc.char_span(
                    pred.start,
                    pred.end,
                    alignment_mode="expand"
                )
                for pred in doc_preds
            ]
            for span in doc_pred_spans:
                try:
                    doc._.mentions += (span,)
                except TypeError or ValueError:
                    warnings.warn("Entity couldn't be added.")

    @abstractmethod
    def predict(self, docs: Iterator[Doc], batch_size=None) -> List[List[Span]]:
        """
        Perform the mentions prediction
        :param docs: A list of spacy Document
        :param batch_size: The batch size
        :return:
        """
        pass

    @staticmethod
    def version() -> str:
        return "v1"

    @property
    def require_existing_ner(self) -> bool:
        return False

    @staticmethod
    def _get_serialize_file(path):
        return os.path.join(path, "mentions_extractor.pkl")

    @staticmethod
    def _get_config_file(path):
        path = os.path.join(path, "mentions_extractor.json")
        path = ensure_path(path)
        return path

    @classmethod
    def from_disk(cls, path, exclude=()):
        serialize_file = cls._get_serialize_file(path)
        with open(serialize_file, "rb") as f:
            return pkl.load(f)

    def to_disk(self, path):
        serialize_file = self._get_serialize_file(path)
        with open(serialize_file, "wb") as f:
            return pkl.dump(self, f)

    def __hash__(self):
        self_repr = f"{self.__class__.__name__}.{self.version()}.{str(self.__dict__)}"
        return zlib.crc32(self_repr.encode())



import os

import spacy
from tqdm import tqdm

from zshot.evaluation.dataset.med_mentions.entities import MEDMENTIONS_TYPE_INV


class Token(object):
    def __init__(self, word, label, label_id):
        self.word = word
        self.label = label
        self.label_id = label_id


def convert_to_iob(id_, text, entities, nlp, end_index, start_index):
    count = 0
    sentences = []
    current_sent_pos = 0
    doc = nlp(text)
    for sent in doc.sents:
        sentence = []
        for tok in sent:
            word = tok.text
            ent = entities[0][1] if entities else "O"
            if entities and tok.idx + current_sent_pos == start_index[0]:
                count += 1
                label = 'B-' + MEDMENTIONS_TYPE_INV[ent]
                label_id = 'B-' + ent
                token = Token(word=word, label=label, label_id=label_id)
            elif entities and start_index[0] < tok.idx + current_sent_pos <= end_index[0]:
                label = 'I-' + MEDMENTIONS_TYPE_INV[ent]
                label_id = 'I-' + ent
                token = Token(word=word, label=label, label_id=label_id)
            else:
                token = Token(word=word, label='O', label_id="O")
            if entities and len(word) + tok.idx + current_sent_pos >= end_index[0]:
                start_index.pop(0)
                end_index.pop(0)
                entities.pop(0)
            sentence.append(token)
        current_sent_pos += len(sent) + 1
        sentences.append((id_, sentence))
    return sentences, count


def preprocess_medmentions(input_path):
    data_path = os.path.join(input_path, 'corpus_pubtator.txt')
    train_id_path = os.path.join(input_path, 'corpus_pubtator_pmids_train.txt')
    dev_id_path = os.path.join(input_path, 'corpus_pubtator_pmids_dev.txt')
    test_id_path = os.path.join(input_path, 'corpus_pubtator_pmids_test.txt')

    nlp = spacy.load("en_core_web_sm")
    sentences = []
    with open(data_path, 'r') as f:
        data = f.readlines()
        ids = []
        titles = []
        abstracts = []
        entities_title = []
        entities_abstract = []
        ends_title = []
        ends_abstract = []
        starts_title = []
        starts_abstract = []
        id_tmp = None
        starts_title_tmp = []
        starts_abstract_tmp = []
        ends_title_tmp = []
        ends_abstract_tmp = []
        entities_title_tmp = []
        entities_abstract_tmp = []
        for line in data:
            if '|t|' in line:
                id_tmp = line.split("|t|")[0]
                ids.append(id_tmp)
                titles.append(line.split("|t|")[1])
            elif '|a|' in line:
                abstracts.append(line.split('|a|')[1])
            elif id_tmp in line:
                _, start_idx, end_idx, ent, label, _ = line.split('\t')
                if len(ends_title_tmp) > 0 and int(end_idx) <= ends_title_tmp[-1] or \
                        len(ends_abstract_tmp) > 0 and int(end_idx) - len(titles[-1]) <= ends_abstract_tmp[-1]:
                    continue
                if int(end_idx) < len(titles[-1]):
                    starts_title_tmp.append(int(start_idx))
                    ends_title_tmp.append(int(end_idx))
                    entities_title_tmp.append((ent, label))
                else:
                    starts_abstract_tmp.append(int(start_idx) - len(titles[-1]))
                    ends_abstract_tmp.append(int(end_idx) - len(titles[-1]))
                    entities_abstract_tmp.append((ent, label))
            else:
                starts_title.append(starts_title_tmp)
                ends_title.append(ends_title_tmp)
                entities_title.append(entities_title_tmp)
                starts_abstract.append(starts_abstract_tmp)
                ends_abstract.append(ends_abstract_tmp)
                entities_abstract.append(entities_abstract_tmp)
                starts_title_tmp = []
                starts_abstract_tmp = []
                ends_title_tmp = []
                ends_abstract_tmp = []
                entities_title_tmp = []
                entities_abstract_tmp = []
                id_tmp = None

        count = 0
        for id_, title, entities_title_tmp, start_title, end_title, abstract, entities_abstract_tmp, \
            start_abstract, end_abstract in tqdm(zip(ids, titles, entities_title, starts_title, ends_title,
                                                     abstracts, entities_abstract, starts_abstract, ends_abstract)):
            sentences_title, count_t = convert_to_iob(id_, title, entities_title_tmp, nlp,
                                                      end_title, start_title)
            sentences_abstract, count_a = convert_to_iob(id_, abstract, entities_abstract_tmp, nlp,
                                                         end_abstract, start_abstract)
            count += count_a + count_t
            sentences += sentences_title
            sentences += sentences_abstract

    with open(train_id_path, "r") as f:
        train_ids = [line.strip() for line in f.readlines()]
        train_sentences = [sent for id_, sent in sentences if id_ in train_ids]
    with open(dev_id_path, "r") as f:
        dev_ids = [line.strip() for line in f.readlines()]
        dev_sentences = [sent for id_, sent in sentences if id_ in dev_ids]
    with open(test_id_path, "r") as f:
        test_ids = [line.strip() for line in f.readlines()]
        test_sentences = [sent for id_, sent in sentences if id_ in test_ids]

    return train_sentences, dev_sentences, test_sentences



import os
import zipfile
from typing import List, Tuple

import torch
from transformers import BertTokenizerFast

from zshot.utils.data_models import Entity
from zshot.utils.data_models import Span
from zshot.linker.smxm.model import BertTaggerMultiClass
from zshot.utils import download_file

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class SmxmInput(dict):
    def __init__(
            self,
            input_ids: torch.Tensor,
            attention_mask: torch.Tensor,
            token_type_ids: torch.Tensor,
            sep_index: torch.Tensor,
            seq_mask: torch.Tensor,
            split: torch.Tensor,
    ):
        config = {
            "input_ids": input_ids.to(device),
            "attention_mask": attention_mask.to(device),
            "token_type_ids": token_type_ids.to(device),
            "sep_index": sep_index.to(device),
            "seq_mask": seq_mask.to(device),
            "split": split.to(device),
        }
        super().__init__(**config)


def load_model(url: str, output_path: str, folder_name: str) -> BertTaggerMultiClass:
    filename = url.rsplit("/", 1)[1]
    model_zipfile_path = os.path.join(output_path, filename)
    model_folder_path = os.path.join(output_path, folder_name)

    if not os.path.isdir(model_folder_path):
        download_file(url, output_path)
        with zipfile.ZipFile(model_zipfile_path, "r") as model_zip:
            model_zip.extractall(output_path)
        os.remove(model_zipfile_path)

    model = BertTaggerMultiClass.from_pretrained(
        model_folder_path, output_hidden_states=True
    ).to(device)

    return model


def predictions_to_span_annotations(
        sentences: List[str],
        predictions: List[List[int]],
        probabilities: List[List[List[float]]],
        entities: List[str],
        tokenizer: BertTokenizerFast,
        max_sentence_tokens: int,
) -> List[List[Span]]:
    span_annotations = []

    for i, sentence in enumerate(sentences):
        sentence_span_annotations = []

        tokenization = tokenizer.encode_plus(
            sentence,
            return_token_type_ids=False,
            return_attention_mask=False,
            return_offsets_mapping=True,
        )

        num_tokens_full_sentence = len(tokenization["input_ids"]) - 2
        token_overflow = num_tokens_full_sentence - max_sentence_tokens
        truncation_offset = token_overflow * (token_overflow > 0)

        offset_mapping = tokenization["offset_mapping"]
        mapping_input_id_to_word = tokenization.encodings[0].word_ids
        words_offset_mappings = {}
        for j, w in enumerate(mapping_input_id_to_word):
            if w in words_offset_mappings:
                words_offset_mappings[w] = (
                    words_offset_mappings[w][0],
                    offset_mapping[j][1],
                )
            elif w is not None:
                words_offset_mappings[w] = offset_mapping[j]

        for j, input_id in enumerate(mapping_input_id_to_word[truncation_offset:-1]):
            pred = predictions[i][j]
            if (
                    entities[pred] != "NEG"
            ) and (
                    input_id is not None
            ) and (
                    (j == 0) or (input_id != mapping_input_id_to_word[j - 1])
            ) and (
                    (words_offset_mappings[input_id][1] - words_offset_mappings[input_id][0]) > 1
            ):
                if (
                        sentence_span_annotations and (sentence_span_annotations[-1].label == entities[pred])
                ) and (
                        sentence_span_annotations[-1].end == words_offset_mappings[input_id][0] - 1
                ):
                    sentence_span_annotations[-1].end = words_offset_mappings[input_id][1]
                    sentence_span_annotations[-1].score = max(sentence_span_annotations[-1].score,
                                                              probabilities[i][j][pred])
                else:
                    sentence_span_annotations.append(
                        Span(words_offset_mappings[input_id][0], words_offset_mappings[input_id][1],
                             entities[pred], probabilities[i][j][pred])
                    )
        span_annotations.append(sentence_span_annotations)

    return span_annotations


def get_entities_names_descriptions(
        entities: List[Entity],
) -> Tuple[List[str], List[str]]:
    if "NEG" in [e.name for e in entities]:
        neg_index = [e.name for e in entities].index("NEG")
        entities.insert(0, entities.pop(neg_index))
    else:
        neg_ent = Entity(
            name="NEG",
            description="Coal, water, oil, etc. are normally used for traditional electricity generation. "
                        "However using liquefied natural gas as fuel for joint circulatory electircity generation has advantages. "
                        "The chief financial officer is the only one there taking the fall. It has a very talented team, eh. "
                        "What will happen to the wildlife? I just tell them, you've got to change. They're here to stay. "
                        "They have no insurance on their cars. What else would you like? Whether holding an international cultural event "
                        "or setting the city's cultural policies, she always asks for the participation or input of other cities and counties.",
        )
        entities.insert(0, neg_ent)

    entity_labels = [e.name for e in entities]
    entity_descriptions = [e.description for e in entities]

    return entity_labels, entity_descriptions



from typing import List


class Trie(object):
    def __init__(self, sequences: List[List[int]] = []):
        self.trie_dict = {}
        for sequence in sequences:
            self.add(sequence)

    def add(self, sequence: List[int]):
        trie = self.trie_dict
        for idx in sequence:
            if idx not in trie:
                trie[idx] = {}
            trie = trie[idx]

    def postfix(self, prefix_sequence: List[int]):
        if len(prefix_sequence) == 1:
            return list(self.trie_dict.keys())
        trie = self.trie_dict
        for pfx in prefix_sequence[1:]:
            if pfx not in trie:
                return []
            trie = trie[pfx]
        return list(trie.keys())



from typing import Iterator, Optional, Union, List

from spacy.tokens import Doc
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM

from zshot.linker.linker import Linker
from zshot.linker.linker_regen.trie import Trie
from zshot.linker.linker_regen.utils import create_input
from zshot.utils.data_models import Entity, Span

MODEL_NAME = "gabriele-picco/regen-disambiguation"

START_ENT_TOKEN = "[START_ENT]"
END_ENT_TOKEN = "[END_ENT]"


class LinkerRegen(Linker):
    """ REGEN linker """
    def __init__(self, max_input_len=384, max_output_len=15, num_beams=10):
        """
        :param max_input_len: Max length of input
        :param max_output_len: Max length of output
        :param num_beams: Number of beans to use
        """
        super().__init__()
        self.model = None
        self.tokenizer = None
        self.trie = None
        self.max_input_len = max_input_len
        self.max_output_len = max_output_len
        self.num_beams = num_beams

    def set_kg(self, entities: Iterator[Entity]):
        """ Set new entities

        :param entities: New entities to use
        """
        super().set_kg(entities)
        self.load_tokenizer()
        self.trie = Trie(
            [
                self.tokenizer(e.name, return_tensors="pt")['input_ids'][0].tolist()
                for e in entities
            ]
        )

    def load_models(self):
        """ Load Model """
        if self.model is None:
            self.model = AutoModelForSeq2SeqLM.from_pretrained(MODEL_NAME)
        self.load_tokenizer()

    def load_tokenizer(self):
        """ Load Tokenizer"""
        if self.tokenizer is None:
            self.tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME, model_max_length=1024)

    def restrict_decode_vocab(self, _, prefix_beam):
        """ Restrict the posibilities of the Beam search to force the text generation """
        return self.trie.postfix(prefix_beam.tolist())

    def predict(self, docs: Iterator[Doc], batch_size: Optional[Union[int, None]] = None) -> List[List[Span]]:
        """
        Perform the entity prediction
        :param docs: A list of spacy Document
        :param batch_size: The batch size
        :return: List Spans for each Document in docs
        """
        self.load_models()
        data_to_link = []
        docs = list(docs)
        for doc_id, doc in enumerate(docs):
            for mention_id, mention in enumerate(doc._.mentions):
                left_context = doc.text[:mention.start_char]
                right_context = doc.text[mention.end_char:]
                sentence = f"{left_context} {START_ENT_TOKEN} {mention.text} {END_ENT_TOKEN} {right_context}"
                data_to_link.append(
                    {
                        "id": doc_id,
                        "mention_id": mention_id,
                        "text": sentence,
                    })

        sentences = [create_input(d['text'],
                                  max_length=self.max_input_len,
                                  start_delimiter=START_ENT_TOKEN,
                                  end_delimiter=END_ENT_TOKEN,
                                  ) for d in data_to_link]
        input_args = {
            k: v
            for k, v in self.tokenizer.batch_encode_plus(
                sentences, padding=True, return_tensors="pt"
            ).items()
        }

        outputs = self.model.generate(
            **input_args,
            min_length=0,
            max_length=self.max_output_len,
            num_beams=self.num_beams,
            num_return_sequences=1,
            output_scores=True,
            return_dict_in_generate=True,
            prefix_allowed_tokens_fn=None
            if self.trie is None
            else self.restrict_decode_vocab,
        )

        docs_pred = {}

        for data, out, score in zip(data_to_link, outputs.sequences, outputs.sequences_scores):
            doc_id = data['id']
            mention = docs[doc_id]._.mentions[data['mention_id']]
            label = self.tokenizer.decode(out, skip_special_tokens=True)
            if doc_id not in docs_pred:
                docs_pred[doc_id] = []
            docs_pred[doc_id].append(Span(mention.start_char, mention.end_char, label=label,
                                          score=score.detach().numpy().tolist()))
        return [val for key, val in sorted(docs_pred.items(), reverse=False)]



import spacy

from zshot.mentions_extractor import MentionsExtractorSpacy
from zshot.mentions_extractor.mentions_extractor_spacy import ExtractorType
from zshot.tests.config import EX_DOCS
from zshot import PipelineConfig


def test_spacy_ner_mentions_extractor():
    nlp = spacy.load("en_core_web_sm")

    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorSpacy(ExtractorType.NER))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names and "ner" in nlp.pipe_names

    doc = nlp(EX_DOCS[1])
    assert doc.ents == ()
    assert len(doc._.mentions) > 0


def test_custom_spacy_mentions_extractor():
    nlp = spacy.load("en_core_web_sm")

    custom_component = MentionsExtractorSpacy(ExtractorType.NER)
    config_zshot = PipelineConfig(mentions_extractor=custom_component, disable_default_ner=False)
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names and "ner" in nlp.pipe_names

    doc = nlp(EX_DOCS[1])
    assert doc.ents == ()
    assert len(doc._.mentions) > 0


def test_spacy_pos_mentions_extractor():
    nlp = spacy.load("en_core_web_sm")

    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorSpacy(ExtractorType.POS))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names and "ner" not in nlp.pipe_names
    doc = nlp(EX_DOCS[1])
    assert doc.ents == ()
    assert len(doc._.mentions) > 0


def test_spacy_ner_mentions_extractor_pipeline():
    nlp = spacy.load("en_core_web_sm")

    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorSpacy(ExtractorType.NER))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names and "ner" in nlp.pipe_names
    docs = [doc for doc in nlp.pipe(EX_DOCS)]
    assert all(doc.ents == () for doc in docs)
    assert all(len(doc._.mentions) > 0 for doc in docs)


def test_spacy_pos_mentions_extractor_pipeline():
    nlp = spacy.load("en_core_web_sm")

    config_zshot = PipelineConfig(mentions_extractor=MentionsExtractorSpacy(ExtractorType.POS))
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names and "ner" not in nlp.pipe_names
    docs = [doc for doc in nlp.pipe(EX_DOCS)]
    assert all(doc.ents == () for doc in docs)
    assert all(len(doc._.mentions) > 0 for doc in docs)



from spacy.tokens import Doc, Span


class Span:
    def __init__(self, start: int, end: int, label: str = None, score: float = None, kb_id: str = None):
        self.start = start
        self.end = end
        self.label = label
        self.score = score
        self.kb_id = kb_id

    def __repr__(self) -> str:
        return f"{self.label}, {self.start}, {self.end}, {self.score}"

    def to_spacy_span(self, doc: Doc) -> Span:
        kwargs = {
            'alignment_mode': 'expand'
        }
        if self.kb_id:
            kwargs.update({'kb_id': self.kb_id})
        if self.label:
            kwargs.update({'label': self.label})

        return doc.char_span(self.start, self.end, **kwargs)

    @staticmethod
    def from_spacy_span(spacy_span: Span, score=None):
        return Span(spacy_span.start_char, spacy_span.end_char, spacy_span.label_, score=score, kb_id=spacy_span.kb_id)



import spacy

from zshot import PipelineConfig, displacy
from zshot.tests.config import EX_ENTITIES, EX_DOCS
from zshot.tests.linker.test_linker import DummyLinkerEnd2End
from zshot.tests.mentions_extractor.test_mention_extractor import DummyMentionsExtractor
from zshot.utils.data_models import Span
from zshot.utils.alignment_utils import align_spans, AlignmentMode, filter_overlapping_spans


def test_alignment_expand():
    tokens = ["I", "am", "going"]
    spans = [Span(2, 4, label="SBJ"), Span(5, 6, label="LOC")]
    tokens_offset = [(0, 1), (2, 4), (5, 10)]
    alignment = align_spans(spans, tokens, tokens_offsets=tokens_offset, alignment_mode=AlignmentMode.expand)
    assert alignment[0] == []
    assert alignment[1] == [0]
    assert alignment[2] == [1]
    assert [spans[a[0]].label if a else "O" for a in alignment] == ["O", "SBJ", "LOC"]


def test_alignment_expand_join_by():
    tokens = ["I", "am", "going"]
    spans = [Span(2, 4, label="SBJ"), Span(5, 6, label="LOC")]
    alignment = align_spans(spans, tokens, join_by=" ", alignment_mode=AlignmentMode.expand)
    assert alignment[0] == []
    assert alignment[1] == [0]
    assert alignment[2] == [1]
    assert [spans[a[0]].label if a else "O" for a in alignment] == ["O", "SBJ", "LOC"]


def test_alignment_expand_2():
    tokens = ["I", "am", "going", "home"]
    spans = [Span(0, 1, label="SBJ"), Span(11, 15, label="LOC")]
    alignment = align_spans(spans, tokens, join_by=" ", alignment_mode=AlignmentMode.expand)
    assert alignment[0] == [0]
    assert alignment[1] == []
    assert alignment[2] == []
    assert alignment[3] == [1]
    assert [spans[a[0]].label if a else "O" for a in alignment] == ["SBJ", "O", "O", "LOC"]


def test_alignment_expand_3():
    """
    text: -- -- --
    s_0:  AA A          0.7
    s_1:  BB             0.5
    s_2:     CC CC     0.9
    s_3:        DD     0.3

    i_*  [0,1] [0,2] [2,3]
    """
    tokens = ["--", "--", "--"]
    spans = [Span(0, 4, label="A", score=0.7),
             Span(0, 2, label="B", score=0.5),
             Span(3, 8, label="C", score=0.9),
             Span(6, 8, label="D", score=0.4)]

    alignment = align_spans(spans, tokens, join_by=" ", alignment_mode=AlignmentMode.expand)

    assert alignment[0] == [0, 1]
    assert alignment[1] == [0, 2]
    assert alignment[2] == [2, 3]


def test_alignment_expand_4():
    """
    text: -- -- --
    s_0:  AA           0.7
    s_1:  BB           0.5
    s_2:     CC CC     0.9
    s_3:        DD     0.3

    i_*  [0,1] [0,2] [2,3]
    """
    tokens = ["--", "--", "--"]
    spans = [Span(0, 3, label="A", score=0.7),
             Span(0, 2, label="B", score=0.5),
             Span(3, 8, label="C", score=0.9),
             Span(6, 8, label="D", score=0.4)]

    alignment = align_spans(spans, tokens, join_by=" ", alignment_mode=AlignmentMode.contract)

    assert alignment[0] == [0, 1]
    assert alignment[1] == [2]
    assert alignment[2] == [2, 3]


def test_alignment_contract():
    tokens = ["I", "am", "going", "home"]
    spans = [Span(1, 3, label="SBJ"), Span(11, 15, label="LOC")]
    alignment = align_spans(spans, tokens, join_by=" ", alignment_mode=AlignmentMode.contract)
    assert alignment[0] == []
    assert alignment[1] == []
    assert alignment[2] == []
    assert alignment[3] == [1]
    assert [spans[a[0]].label if a else "O" for a in alignment] == ["O", "O", "O", "LOC"]


def test_alignment_expand_overlaps():
    """
    text: -- -- --
    s_0:  AA A          0.7
    s_1:  BB             0.5
    s_2:     CC CC     0.9
    s_3:        DD     0.3

    i_*  [0,1] [0,2] [2,3]
    """
    tokens = ["--", "--", "--"]
    spans = [Span(0, 4, label="A", score=0.7),
             Span(0, 2, label="B", score=0.5),
             Span(3, 8, label="C", score=0.9),
             Span(6, 8, label="D", score=0.4)]

    filtered_spans = filter_overlapping_spans(spans, tokens, join_by=" ", alignment_mode=AlignmentMode.expand)

    assert len(filtered_spans) == 2
    assert filtered_spans[0].start == 0 and filtered_spans[0].end == 2
    assert filtered_spans[0].label == "A"
    assert filtered_spans[1].start == 3 and filtered_spans[1].end == 8
    assert filtered_spans[1].label == "C"


def test_alignment_expand_overlaps_2():
    """
    text: -- -- --
    s_0:  AA           0.7
    s_1:  BB           0.5
    s_2:     CC CC     0.9
    s_3:        DD     0.3

    i_*  [0,1] [0,2] [2,3]
    """
    tokens = ["--", "--", "--"]
    spans = [Span(0, 2, label="A", score=0.7),
             Span(0, 2, label="B", score=0.5),
             Span(3, 8, label="C", score=0.9),
             Span(6, 8, label="D", score=0.4)]

    filtered_spans = filter_overlapping_spans(spans, tokens, join_by=" ", alignment_mode=AlignmentMode.expand)

    assert len(filtered_spans) == 2
    assert filtered_spans[0].start == 0 and filtered_spans[0].end == 2
    assert filtered_spans[0].label == "A"
    assert filtered_spans[1].start == 3 and filtered_spans[1].end == 8
    assert filtered_spans[1].label == "C"


def test_alignment_expand_overlaps_no_score():
    """
    text: -- -- --
    s_0:  AA           0.7
    s_1:  BB           0.5
    s_2:     CC CC     0.9
    s_3:        DD     0.3

    i_*  [0,1] [0,2] [2,3]
    """
    tokens = ["--", "--", "--"]
    spans = [Span(0, 2, label="A", score=0.7),
             Span(0, 2, label="B"),
             Span(3, 8, label="C"),
             Span(6, 8, label="D", score=0.4)]

    filtered_spans = filter_overlapping_spans(spans, tokens, join_by=" ", alignment_mode=AlignmentMode.expand)

    assert len(filtered_spans) == 2
    assert filtered_spans[0].start == 0 and filtered_spans[0].end == 2
    assert filtered_spans[0].label == "A"
    assert filtered_spans[1].start == 3 and filtered_spans[1].end == 8
    assert filtered_spans[1].label == "C"


def test_displacy_render():
    nlp = spacy.blank("en")

    nlp.add_pipe("zshot", config=PipelineConfig(
        mentions_extractor=DummyMentionsExtractor(),
        linker=DummyLinkerEnd2End(),
        entities=EX_ENTITIES), last=True)
    doc = nlp(EX_DOCS[1])
    assert len(doc.ents) > 0
    assert len(doc._.spans) > 0
    res = displacy.render(doc, style="ent", jupyter=False)
    assert res is not None



from typing import Optional, Union, Dict, List

import spacy

from zshot.utils.data_models import Entity
from zshot.linker import Linker
from zshot.mentions_extractor import MentionsExtractor


class PipelineConfig(dict):

    def __init__(self,
                 mentions_extractor: Optional[MentionsExtractor] = None,
                 linker: Optional[Union[Linker, str]] = None,
                 entities: Optional[Union[Dict[str, str], List[Entity], List[str], str]] = None,
                 disable_default_ner: Optional[bool] = True) -> None:
        config = {}

        if mentions_extractor:
            mention_extractor_id = PipelineConfig.param(mentions_extractor)
            config.update({'mentions_extractor': mention_extractor_id})

        if linker:
            linker_id = PipelineConfig.param(linker)
            config.update({'linker': linker_id})

        if entities:
            entities_id = PipelineConfig.param(entities)
            config.update({'entities': entities_id})

        if disable_default_ner:
            config.update({'disable_default_ner': disable_default_ner})

        super().__init__(**config)

    @staticmethod
    def param(param) -> str:
        if isinstance(param, list):
            instance_hash = hash(hash(param[0]) + hash(param[-1]))
        else:
            instance_hash = hash(param)

        instance_id = f"zshot.{param.__class__.__name__}.{instance_hash}"

        @spacy.registry.misc(instance_id)
        def create_custom_component():
            return param

        return instance_id



import os
import shutil

import spacy

from zshot import PipelineConfig
from zshot.config import MODELS_CACHE_PATH
from zshot.linker import LinkerSMXM
from zshot.linker.linker_smxm import SMXM_MODEL_FILES_URL, SMXM_MODEL_FOLDER_NAME
from zshot.linker.smxm.model import BertTaggerMultiClass
from zshot.linker.smxm.utils import load_model
from zshot.tests.config import EX_DOCS, EX_ENTITIES


def test_smxm_download():
    model_folder_path = os.path.join(MODELS_CACHE_PATH, SMXM_MODEL_FOLDER_NAME)
    if os.path.exists(model_folder_path):
        shutil.rmtree(model_folder_path)

    model = load_model(SMXM_MODEL_FILES_URL, MODELS_CACHE_PATH, SMXM_MODEL_FOLDER_NAME)

    assert isinstance(model, BertTaggerMultiClass)


def test_smxm_linker():
    nlp = spacy.blank("en")
    smxm_config = PipelineConfig(
        linker=LinkerSMXM(),
        entities=EX_ENTITIES
    )
    nlp.add_pipe("zshot", config=smxm_config, last=True)
    assert "zshot" in nlp.pipe_names

    doc = nlp(EX_DOCS[1])
    assert len(doc.ents) > 0


def test_smxm_linker_pipeline():
    nlp = spacy.blank("en")
    smxm_config = PipelineConfig(
        linker=LinkerSMXM(),
        entities=EX_ENTITIES
    )
    nlp.add_pipe("zshot", config=smxm_config, last=True)
    assert "zshot" in nlp.pipe_names

    docs = [doc for doc in nlp.pipe(EX_DOCS)]
    assert all(len(doc.ents) > 0 for doc in docs)


def test_smxm_linker_no_entities():
    nlp = spacy.blank("en")
    smxm_config = PipelineConfig(
        linker=LinkerSMXM(),
        entities=[]
    )
    nlp.add_pipe("zshot", config=smxm_config, last=True)
    assert "zshot" in nlp.pipe_names

    doc = nlp(EX_DOCS[1])
    assert len(doc.ents) == 0



def create_input(sentence, max_length, start_delimiter, end_delimiter):
    sent_list = sentence.split(" ")
    if len(sent_list) < max_length:
        return sentence
    else:
        end_delimiter_index = sent_list.index(end_delimiter)
        start_delimiter_index = sent_list.index(start_delimiter)
        half_context = (max_length - (end_delimiter_index - start_delimiter_index)) // 2
        left_index = max(0, start_delimiter_index - half_context)
        right_index = min(len(sent_list), end_delimiter_index + half_context + (
            half_context - (start_delimiter_index - left_index)))
        left_index = left_index - max(0, (half_context - (right_index - end_delimiter_index)))
        print(len(sent_list[left_index:right_index]))
        return " ".join(sent_list[left_index:right_index])



from typing import Any, Dict, List, Tuple, Union

import torch
from torch.utils.data import Dataset
from transformers import BertTokenizerFast

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class ByDescriptionTaggerDataset(Dataset):
    def __init__(self, data: List[Dict[str, Any]]):
        self.data = data

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: Any) -> List[Dict[str, Any]]:
        if torch.is_tensor(idx):
            idx = idx.tolist()
        return self.data[idx]


def encode_data(
    sentences: List[str],
    entity_labels: List[str],
    entity_descriptions: List[str],
    tokenizer: BertTokenizerFast,
) -> Tuple[List[Dict[str, Any]], int]:
    encoded_data = []
    tokenized_descriptions = [
        tokenizer.tokenize(description) for description in entity_descriptions
    ]
    max_descriptions_tokens = max([len(d) for d in tokenized_descriptions])
    max_sentence_tokens = 512 - max_descriptions_tokens - 3

    for sentence in sentences:
        tokenized_sentence = tokenizer.tokenize(
            sentence, truncation=True, max_length=max_sentence_tokens
        )
        tokenized_texts_list = []
        input_ids_list = []
        input_masks_list = []
        segment_ids_list = []

        for tokenized_description in tokenized_descriptions:
            combined_tokenized_text = (
                ["[CLS]"] + tokenized_sentence + ["[SEP]"] + tokenized_description + ["[SEP]"]
            )
            split_index = combined_tokenized_text.index("[SEP]")
            input_ids = tokenizer.convert_tokens_to_ids(combined_tokenized_text)
            input_mask = [1] * len(input_ids)
            segment_ids = ([0] * split_index) + (
                [1] * (len(combined_tokenized_text) - split_index)
            )

            tokenized_texts_list.append(combined_tokenized_text)
            input_ids_list.append(input_ids)
            input_masks_list.append(input_mask)
            segment_ids_list.append(segment_ids)

        encoded_data.append(
            {
                "text": tokenized_texts_list,
                "input_ids": input_ids_list,
                "input_masks": input_masks_list,
                "segment_ids": segment_ids_list,
                "sep_index": split_index,
            }
        )

    return encoded_data, max_sentence_tokens


def tagger_multiclass_collator(
    data: Union[List[Dict[str, Any]], Dict[str, Any]]
) -> Tuple[
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
]:
    if isinstance(data, dict):
        data = [data]

    input_ids_lists = [f["input_ids"] for f in data]
    input_masks_lists = [f["input_masks"] for f in data]
    segment_ids_lists = [f["segment_ids"] for f in data]
    batch_size = len(data)

    sep_index = [f["sep_index"] for f in data]
    max_sep_index = max(sep_index)

    padded_bert_tokens_list = []
    padded_bert_masks_list = []
    padded_bert_segments_list = []

    padded_sequence_mask = []

    sentence_lengths = [
        max([len(bert_tokens) for bert_tokens in input_ids_lists[i]])
        for i in range(len(input_ids_lists))
    ]
    longest_sent = max(sentence_lengths)

    for i in range(batch_size):
        padded_bert_tokens = []
        padded_bert_masks = []
        padded_bert_segments = []

        for j in range(len(input_ids_lists[i])):
            padding = [0] * (longest_sent - len(input_ids_lists[i][j]))
            padded_bert_tokens.append(input_ids_lists[i][j] + padding)
            padded_bert_masks.append(input_masks_lists[i][j] + padding)
            padded_bert_segments.append(segment_ids_lists[i][j] + padding)

        bert_tokens_t = torch.tensor(padded_bert_tokens).to(device)
        bert_masks_t = torch.tensor(padded_bert_masks).to(device)
        bert_segments_t = torch.tensor(padded_bert_segments).to(device)

        padded_bert_tokens_list.append(bert_tokens_t)
        padded_bert_masks_list.append(bert_masks_t)
        padded_bert_segments_list.append(bert_segments_t)

        padded_sequence_mask.append(
            sep_index[i] * [1] + (max_sep_index - sep_index[i]) * [0]
        )

    bert_tokens_t = torch.stack(padded_bert_tokens_list).transpose(1, 0)
    bert_masks_t = torch.stack(padded_bert_masks_list).transpose(1, 0)
    bert_segments_t = torch.stack(padded_bert_segments_list).transpose(1, 0)
    bert_seq_mask_t = torch.tensor(padded_sequence_mask).to(
        device=device, dtype=torch.uint8
    )
    bert_sep_t = torch.tensor(sep_index).to(device=device)

    split = torch.tensor(2).to(device=device)

    return (
        bert_tokens_t,
        bert_masks_t,
        bert_segments_t,
        bert_sep_t,
        bert_seq_mask_t,
        split,
    )



import pkgutil

import spacy

from zshot.linker import LinkerTARS
from zshot.tests.config import EX_DOCS, EX_ENTITIES
from zshot import PipelineConfig

OVERLAP_TEXT = "Senator McConnell in addition to this the New York Times editors wrote in reaction to the " \
               "Supreme Court 's decision striking down the military tribunal set up to private detainees " \
               "being held in Guantanamo bay it is far more than a narrow ruling on the issue of military courts . " \
               "It is an important and welcome reaffirmation That even in times of war the law is what the " \
               "constitution the statuette books and the Geneva convention say it is . " \
               "Not what the President wants it to be /."
INCOMPLETE_SPANS_TEXT = "-LSB- -LSB- They attacked small bridges and small districts , " \
                        "and generally looted these stations . -RSB- -RSB-"


def test_tars_end2end_with_entities():
    if not pkgutil.find_loader("flair"):
        return
    nlp = spacy.blank("en")

    config_zshot = PipelineConfig(linker=LinkerTARS(), entities=EX_ENTITIES)
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])
    assert doc.ents != ()


def test_tars_end2end_pipeline_with_entities():
    if not pkgutil.find_loader("flair"):
        return
    nlp = spacy.blank("en")

    config_zshot = PipelineConfig(linker=LinkerTARS(), entities=EX_ENTITIES)
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    docs = [doc for doc in nlp.pipe(EX_DOCS)]
    assert all(doc.ents != () for doc in docs)


def test_tars_end2end_overlap():
    if not pkgutil.find_loader("flair"):
        return
    nlp = spacy.blank("en")

    config_zshot = PipelineConfig(linker=LinkerTARS(), entities=["company", "location", "organic compound"])
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    nlp(OVERLAP_TEXT)


def test_tars_end2end_incomplete_spans():
    if not pkgutil.find_loader("flair"):
        return
    nlp = spacy.blank("en")

    config_zshot = PipelineConfig(linker=LinkerTARS())
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(INCOMPLETE_SPANS_TEXT)
    assert len(doc.ents) > 0



from zshot.utils.data_models import Entity

EX_DOCS = ["The Domain Name System (DNS) is the hierarchical and decentralized naming system used to identify"
           " computers, services, and other resources reachable through the Internet or other Internet Protocol"
           " (IP) networks.",
           "International Business Machines Corporation (IBM) is an American multinational technology corporation"
           " headquartered in Armonk, New York, with operations in over 171 countries."]


EX_ENTITIES = \
    [
        Entity(name="apple", description="the apple fruit"),
        Entity(name="DNS", description="domain name system", vocabulary=["DNS", "Domain Name System"]),
        Entity(name="IBM", description="technology corporation", vocabulary=["IBM", "International Business machine"]),
        Entity(name="NYC", description="New York city"),
        Entity(name="Florida", description="southeasternmost U.S. state"),
        Entity(name="Paris", description="Paris is located in northern central France, "
                                         "in a north-bending arc of the river Seine"),
    ]



from zshot.linker.linker_blink import LinkerBlink  # noqa: F401
from zshot.linker.linker_regen.linker_regen import LinkerRegen  # noqa: F401
from zshot.linker.linker import Linker  # noqa: F401
from zshot.linker.linker_smxm import LinkerSMXM  # noqa: F401
from zshot.linker.linker_tars import LinkerTARS  # noqa: F401



import argparse

import spacy

from zshot import PipelineConfig
from zshot.evaluation.metrics.seqeval.seqeval import Seqeval
from zshot.evaluation.zshot_evaluate import evaluate
from zshot.linker import LinkerTARS, LinkerSMXM
from zshot.mentions_extractor import MentionsExtractorSpacy, MentionsExtractorFlair
from zshot.mentions_extractor.utils import ExtractorType

MENTION_EXTRACTORS = {
    "spacy_pos": lambda: MentionsExtractorSpacy(ExtractorType.POS),
    "spacy_ner": lambda: MentionsExtractorSpacy(ExtractorType.NER),
    "flair_pos": lambda: MentionsExtractorFlair(ExtractorType.POS),
    "flair_ner": lambda: MentionsExtractorFlair(ExtractorType.NER),
}
LINKERS = {
    "tars": LinkerTARS,
    "smxm": LinkerSMXM
}
END2END = ['tars', 'smxm']

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset", default="ontonotes", type=str, help="Name or path to the validation data. Comma separated")
    parser.add_argument("--splits", required=False, default="train, test, validation", type=str,
                        help="Splits to evaluate. Comma separated")
    parser.add_argument("--mode", required=False, default="full", type=str,
                        help="Evaluation mode. One of: full; mentions_extractor; linker")
    parser.add_argument("--mentions_extractor", required=False, default="all", type=str,
                        help="Mentions extractor to evaluate. One of: all; spacy_pos; spacy_ner; flair_pos; flair_ner")
    parser.add_argument("--linker", required=False, default="all", type=str,
                        help="Linker to evaluate. One of: all; tars")
    args = parser.parse_args()

    args.splits = args.splits.split(",")
    args.dataset = args.dataset.split(",")

    configs = {}
    if args.mentions_extractor == "all":
        mentions_extractors = list(MENTION_EXTRACTORS.keys())
    else:
        mentions_extractors = [args.mentions_extractor]

    if args.linker == "all":
        linkers = list(LINKERS.keys())
    else:
        linkers = [args.linker]

    if args.mode == "full":
        if linkers:
            for linker in linkers:
                if linker not in END2END:
                    for mentions_extractor in mentions_extractors:
                        configs[f"{mentions_extractor}_{linker}"] = PipelineConfig(
                            mentions_extractor=MENTION_EXTRACTORS[mentions_extractor](),
                            linker=LINKERS[linker]()
                        )
                else:
                    configs[linker] = PipelineConfig(linker=LINKERS[linker]())
        for mentions_extractor in mentions_extractors:
            configs[mentions_extractor] = PipelineConfig(mentions_extractor=MENTION_EXTRACTORS[mentions_extractor]())
    elif args.mode == "mentions_extractor":
        for mentions_extractor in mentions_extractors:
            configs[mentions_extractor] = PipelineConfig(mentions_extractor=MENTION_EXTRACTORS[mentions_extractor]())
    elif args.mode == "linker":
        for linker in linkers:
            if linker not in END2END:
                configs[linker] = PipelineConfig(mentions_extractor=MENTION_EXTRACTORS['spacy_ner'](),
                                                 linker=LINKERS[linker]())
            else:
                configs[linker] = PipelineConfig(linker=LINKERS[linker]())

    print(f"Evaluating {args.dataset} with splits {args.splits}")
    for key, config in configs.items():
        print(f"Evaluating {key}")
        nlp = spacy.blank("en") if "spacy" not in key else spacy.load("en_core_web_sm")
        nlp.add_pipe("zshot", config=config, last=True)

        print(evaluate(nlp, args.dataset, splits=args.splits, metric=Seqeval()))



import pkgutil
from typing import Iterator, Optional, Union, List

from spacy.tokens.doc import Doc

from zshot.utils.data_models import Entity, Span
from zshot.linker.linker import Linker


class LinkerTARS(Linker):
    """ TARS end2end Linker """
    def __init__(self, default_entities: Optional[str] = "conll-short"):
        """
        :param default_entities: Default entities to use in case no custom ones are set
        One of:
            - 'conll-short'
            - 'ontonotes-long'
            - 'ontonotes-short'
            - 'wnut_17-long'
            - 'wnut_17-short'
        """
        super().__init__()
        if not pkgutil.find_loader("flair"):
            raise Exception("Flair module not installed. You need to install Flair for using this class."
                            "Install it with: pip install flair==0.11")

        self.is_end2end = True
        self.default_entities = default_entities
        self.model = None
        self.task = None

    def set_kg(self, entities: Iterator[Entity]):
        """ Set new entities in the model

        :param entities: New entities to use
        """
        old_entities = self._entities
        super().set_kg(entities)
        if old_entities != entities:
            self.flat_entities()
            self.task = f'zshot.ner.{hash(tuple(self._entities))}'
            if not self.model:
                self.load_models()
            self.model.add_and_switch_to_new_task(self.task,
                                                  self._entities, label_type='ner')

    def flat_entities(self):
        """ As TARS use only the labels, take just the name of the entities and not the description """
        if isinstance(self._entities, dict):
            self._entities = list(self._entities.keys())
        if isinstance(self._entities, list):
            self._entities = [e.name if type(e) == Entity else e for e in self._entities]
        if self._entities is None:
            self._entities = []

    def load_models(self):
        """ Load TARS model if its not initialized"""
        if not self.model:
            from flair.models import TARSTagger
            self.model = TARSTagger.load('tars-ner')

            if not self.entities:
                self.model.switch_to_task(self.default_entities)
                self.task = self.default_entities
            else:
                self.flat_entities()
                self.task = f'zshot.ner.{hash(tuple(self._entities))}'
                self.model.add_and_switch_to_new_task(self.task,
                                                      self._entities, label_type='ner')

    def predict(self, docs: Iterator[Doc], batch_size: Optional[Union[int, None]] = None) -> List[List[Span]]:
        """
        Perform the entity prediction
        :param docs: A list of spacy Document
        :param batch_size: The batch size
        :return: List Spans for each Document in docs
        """
        from flair.data import Sentence

        self.load_models()
        self.model.switch_to_task(self.task)

        sentences = [
            Sentence(str(doc), use_tokenizer=True) for doc in docs
        ]
        kwargs = {'mini_batch_size': batch_size} if batch_size else {}
        self.model.predict(sentences, **kwargs)

        spans_annotations = []
        for sent, doc in zip(sentences, docs):
            sent_mentions = sent.get_spans('ner')
            spans = [
                Span(mention.start_position, mention.end_position, mention.tag, mention.score)
                for mention in sent_mentions
            ]
            spans_annotations.append(spans)

        return spans_annotations



from zshot.utils.data_models import Entity

MEDMENTIONS_TYPE_INV = {
    'T058': "Health_Care_Activity", "T062": "Research_Activity", "T037": "Injury_or_Poisoning",
    "T038": "Biologic_Function", "T005": "Virus", "T007": "Bacterium", "T204": "Eukaryote",
    "T017": "Anotomical_Structure", "T074": "Medical_Device", "T031": "Body_Substance", "T103": "Chemical",
    "T168": "Food", "T201": "Clinical_Attribute", "T033": "Finding", "T082": "Spatial_Concept",
    "T022": "Body_System", "T091": "Biomedical_Occupation_or_Discipline", "T092": "Organization",
    "T097": "Professional_or_Occupational_Group", "T098": "Population_Group", "T170": "Intellectual_Product",
    "NEG": "NEG"
}

MEDMENTIONS_SPLITS = {
    "train": ['Biologic_Function', 'Chemical', 'Health_Care_Activity', 'Anotomical_Structure', "Finding",
              "Spatial_Concept", "Intellectual_Product", "Research_Activity", 'Medical_Device', 'Eukaryote',
              'Population_Group'],
    "validation": ['Biomedical_Occupation_or_Discipline', 'Virus', 'Clinical_Attribute', 'Injury_or_Poisoning',
                   'Organization'],
    "test": ['Body_System', 'Food', 'Body_Substance', 'Bacterium', 'Professional_or_Occupational_Group']}

MEDMENTIONS_ENTITIES = [
    Entity(name="T058", description="Healthcare activity"),
    Entity(name="T062", description="Research activity"),
    Entity(name="T037", description="Injury or Poisoning"),
    Entity(name="T038", description="Biologic Function"),
    Entity(name="T005", description="Virus"),
    Entity(name="T007", description="Bacterium"),
    Entity(name="T204", description="Eukaryote"),
    Entity(name="T017", description="Anotomical structure"),
    Entity(name="T074", description="Medical device"),
    Entity(name="T031", description="Body substance"),
    Entity(name="T103", description="Chemical"),
    Entity(name="T168", description="Food"),
    Entity(name="T201", description="Clinical Attribute"),
    Entity(name="T033", description="Finding"),
    Entity(name="T082", description="Spatial Concept"),
    Entity(name="T022", description="Body System"),
    Entity(name="T091", description="Biomedical occupation or discipline"),
    Entity(name="T092", description="Organization"),
    Entity(name="T097", description="Professional or occupational group"),
    Entity(name="T098", description="Population group"),
    Entity(name="T170", description="Intellectual product")
]



from typing import Iterator

import spacy
from spacy.tokens import Doc

from zshot import Linker, PipelineConfig
from zshot.utils.data_models import Span
from zshot.tests.config import EX_DOCS, EX_ENTITIES
from zshot.tests.mentions_extractor.test_mention_extractor import DummyMentionsExtractor


class DummyLinker(Linker):

    def predict(self, docs: Iterator[Doc], batch_size=None):
        return [
            [Span(mention.start_char, mention.end_char, label='label') for mention in doc._.mentions]
            for doc in docs
        ]


class DummyLinkerEnd2End(Linker):

    @property
    def is_end2end(self) -> bool:
        return True

    def predict(self, docs: Iterator[Doc], batch_size=None):
        return [[Span(0, len(doc.text) - 1, label='label')] for doc in docs]


class DummyLinkerWithEntities(Linker):

    def predict(self, docs: Iterator[Doc], batch_size=None):
        entities = self.entities
        return [
            [
                Span(mention.start_char, mention.end_char, label=entities[idx].name)
                for idx, mention in enumerate(doc._.mentions)
            ]
            for doc in docs
        ]


def test_dummy_linker():
    nlp = spacy.blank("en")
    config = PipelineConfig(
        mentions_extractor=DummyMentionsExtractor(),
        linker=DummyLinker())
    nlp.add_pipe("zshot", config=config, last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])
    assert len(doc._.mentions) > 0
    assert len(doc.ents) > 0
    assert len(doc._.spans) > 0


def test_dummy_linker_with_entities_config():
    nlp = spacy.blank("en")

    nlp.add_pipe("zshot", config=PipelineConfig(
        mentions_extractor=DummyMentionsExtractor(),
        linker=DummyLinker(),
        entities=EX_ENTITIES), last=True)

    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])

    assert len(doc._.mentions) > 0
    assert len(doc.ents) > 0
    assert len(doc._.spans) > 0
    assert all([bool(ent.label_) for ent in doc.ents])


def test_dummy_linker_end2end():
    nlp = spacy.blank("en")

    nlp.add_pipe("zshot", config=PipelineConfig(
        mentions_extractor=DummyMentionsExtractor(),
        linker=DummyLinkerEnd2End(),
        entities=EX_ENTITIES), last=True)

    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[1])

    assert len(doc._.mentions) == 0
    assert len(doc.ents) > 0
    assert len(doc._.spans) > 0



from typing import List, Optional, Dict

import datasets


class DatasetWithEntities(datasets.Dataset):
    # TODO: Implement save and load methods
    def __init__(self, arrow_table: datasets.table.Table,
                 info: Optional[datasets.info.DatasetInfo] = None,
                 split: Optional[datasets.splits.NamedSplit] = None,
                 indices_table: Optional[datasets.table.Table] = None,
                 fingerprint: Optional[str] = None,
                 entities: List[Dict[str, str]] = None):
        super().__init__(arrow_table=arrow_table, info=info, split=split,
                         indices_table=indices_table, fingerprint=fingerprint)
        self.entities = entities

    @classmethod
    def from_dict(
            cls,
            mapping: dict,
            features: Optional[datasets.features.Features] = None,
            info: Optional[datasets.info.DatasetInfo] = None,
            split: Optional[datasets.splits.NamedSplit] = None,
            entities: List[Dict[str, str]] = None
    ) -> "DatasetWithEntities":
        dataset = super().from_dict(mapping=mapping, features=features, info=info, split=split)
        dataset.entities = entities
        return dataset

    def __repr__(self):
        return f"Dataset({{\n    features: {list(self.features.keys())},\n    num_rows: {self.num_rows}," \
               f"\n    entities: {[ent.name for ent in self.entities if self.entities is not None]}\n}})"



import json
import logging
import os
from typing import Dict, Optional, List, Union, Iterator

from catalogue import RegistryError
from spacy.language import Language
from spacy.tokens import Doc
from spacy.util import registry as spacy_registry, ensure_path

from zshot.utils.data_models import Entity
from zshot.linker import Linker
from zshot.mentions_extractor import MentionsExtractor
from zshot.pipeline_config import PipelineConfig


@Language.factory("zshot", default_config={
    "entities": None,
    "mentions_extractor": None,
    "linker": None,
    "disable_default_ner": True
})
def create_zshot_component(nlp: Language, name: str,
                           entities: Optional[Union[Dict[str, str], List[Entity], str]],
                           mentions_extractor: Optional[Union[MentionsExtractor, str]],
                           linker: Optional[Union[Linker, str]],
                           disable_default_ner: Optional[bool] = True):
    return Zshot(nlp, entities, mentions_extractor, linker, disable_default_ner)


class Zshot:

    def __init__(self, nlp: Language,
                 entities,
                 mentions_extractor,
                 linker,
                 disable_default_ner: Optional[bool] = True):
        self.nlp = nlp
        self.entities = entities
        self.mentions_extractor = mentions_extractor
        self.linker = linker
        self.disable_default_ner = disable_default_ner
        self.setup()

    def setup(self):
        # Load Entities from registered function ID if provided
        if isinstance(self.entities, str):
            try:
                self.entities = spacy_registry.get(registry_name='misc', func_name=self.entities)()
            except RegistryError:
                logging.warning(f"Missing entities: {self.entities}")
                self.entities = None
        if isinstance(self.entities, list) and len(self.entities) > 0 and isinstance(self.entities[0], dict):
            self.entities = list(map(lambda e: Entity(**e), self.entities))
        elif isinstance(self.entities, dict):
            self.entities = [Entity(name=name, description=description) for name, description in self.entities.items()]

        # Load Mention Extractor from registered function ID if provided
        if isinstance(self.mentions_extractor, str):
            try:
                self.mentions_extractor = spacy_registry.get(registry_name='misc', func_name=self.mentions_extractor)()
            except RegistryError:
                self.mentions_extractor = None

        # Load Linker from registered function ID if provided
        if isinstance(self.linker, str):
            try:
                self.linker = spacy_registry.get(registry_name='misc', func_name=self.linker)()
            except RegistryError:
                self.linker = None

        if self.mentions_extractor and self.mentions_extractor.require_existing_ner \
                and "ner" not in self.nlp.pipe_names:
            raise ValueError(f"The pipeline you are using does not contains a NER,"
                             f" but mentions extractor {self.mentions_extractor.__class__.__name__} rely on"
                             f"an existing NER")

        if self.mentions_extractor and (self.linker is not None and self.linker.is_end2end):
            logging.warning("Using linker end2end. Disabling mentions_extractor")
            self.mentions_extractor = None

        if self.disable_default_ner and \
                (not self.mentions_extractor or not self.mentions_extractor.require_existing_ner) \
                and "ner" in self.nlp.pipe_names:
            logging.warning("Disabling default NER")
            self.nlp.disable_pipes("ner")

        if not Doc.has_extension("mentions"):
            Doc.set_extension("mentions", default=[])

        if not Doc.has_extension("spans"):
            Doc.set_extension("spans", default=[])

    def __call__(self, doc: Doc) -> Doc:
        # Add the matched spans when doc is processed
        self.extracts_mentions([doc])
        self.link_entities([doc])
        return doc

    def pipe(self, docs: Iterator[Doc], batch_size: int, **kwargs):
        """.
        docs: A sequence of spacy documents.
        YIELDS (Doc): A sequence of Doc objects, in order.
        """
        docs = list(docs)
        self.extracts_mentions(docs, batch_size=batch_size)
        self.link_entities(docs, batch_size=batch_size)
        for doc in docs:
            yield doc

    def extracts_mentions(self, docs: Iterator[Doc], batch_size=None):
        if self.mentions_extractor and not (self.linker is not None and self.linker.is_end2end):
            self.mentions_extractor.set_kg(self.entities)
            self.mentions_extractor.extract_mentions(docs, batch_size=batch_size)

    def link_entities(self, docs, batch_size=None):
        if self.linker:
            self.linker.set_kg(self.entities)
            self.linker.link(docs, batch_size=batch_size)

    def from_disk(self, path, exclude=()):
        with open(os.path.join(path, "config.cfg"), "r") as f:
            config = json.load(f)

        try:
            mentions_extractor = MentionsExtractor.from_disk(path)
        except FileNotFoundError:
            mentions_extractor = None
        else:
            self.mentions_extractor = mentions_extractor

        try:
            linker = Linker.from_disk(path)
        except FileNotFoundError:
            linker = None
        else:
            self.linker = linker

        PipelineConfig(mentions_extractor=mentions_extractor, linker=linker)

        self.disable_default_ner = config['disable_default_ner']
        self.setup()

    def to_disk(self, path, exclude=()):
        path = ensure_path(path)
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        config = {
            "disable_default_ner": self.disable_default_ner
        }

        with open(os.path.join(path, "config.cfg"), "w") as f:
            json.dump(config, f)

        if self.mentions_extractor:
            self.mentions_extractor.to_disk(path)
        if self.linker:
            self.linker.to_disk(path)



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

version = '0.0.2'

setup(name='zshot',
      version=version,
      description="Zero and Few shot named entity recognition",
      long_description_content_type='text/markdown',
      long_description=long_description,
      classifiers=[],
      keywords='NER Zero-Shot Few-Shot',
      author='IBM Research',
      author_email='',
      url='https://ibm.github.io/zshot',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          "spacy>=3.4.1",
          "requests>=2.28",
          "appdata~=2.1.2",
          "tqdm>=4.62.3",
          "setuptools~=60.0.0",  # Needed to install dynamic packages from source (e.g. Blink)
          "prettytable>=3.4",
          "torch>=1",
          "transformers>=4.20",
          "datasets>=2.2.2",
          "evaluate>=0.2.2",
          "seqeval>=1.2.2",
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



from typing import Iterator, List, Optional, Union

import torch
from spacy.tokens import Doc
from torch.utils.data import DataLoader
from transformers import BertTokenizerFast

from zshot.config import MODELS_CACHE_PATH
from zshot.linker.linker import Linker
from zshot.linker.smxm.data import (
    ByDescriptionTaggerDataset,
    encode_data,
    tagger_multiclass_collator
)
from zshot.linker.smxm.utils import (
    SmxmInput,
    get_entities_names_descriptions,
    load_model,
    predictions_to_span_annotations,
)
from zshot.utils.data_models import Span

SMXM_MODEL_FILES_URL = (
    "https://ibm.box.com/shared/static/duni7p7i4gbk0prksc6zv5uahiemfy00.zip"
)
SMXM_MODEL_FOLDER_NAME = "BertTaggerMultiClass_config03_mode_tagger_multiclass_filtered_classes__entity_descriptions_mode_annotation_guidelines__per_gpu_train_batch_size_7/checkpoint"


class LinkerSMXM(Linker):
    """ SMXM linker """

    def __init__(self):
        super().__init__()

        self.tokenizer = BertTokenizerFast.from_pretrained(
            "bert-large-cased", truncation_side="left"
        )

        self.model = None
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    @property
    def is_end2end(self) -> bool:
        """ SMXM is end2end model"""
        return True

    def load_models(self):
        """ Load SMXM model """
        if self.model is None:
            self.model = load_model(
                SMXM_MODEL_FILES_URL, MODELS_CACHE_PATH, SMXM_MODEL_FOLDER_NAME
            )

    def predict(self, docs: Iterator[Doc], batch_size: Optional[Union[int, None]] = None) -> List[List[Span]]:
        """
        Perform the entity prediction
        :param docs: A list of spacy Document
        :param batch_size: The batch size
        :return: List Spans for each Document in docs
        """
        if not self._entities:
            return []

        entity_labels, entity_descriptions = get_entities_names_descriptions(self._entities)
        sentences = [doc.text for doc in docs]

        self.load_models()

        encoded_data, max_sentence_tokens = encode_data(
            sentences, entity_labels, entity_descriptions, self.tokenizer
        )
        dataset = ByDescriptionTaggerDataset(encoded_data)
        dataloader = DataLoader(
            dataset, batch_size=batch_size, collate_fn=tagger_multiclass_collator
        )

        preds = []
        probabilities = []
        for batch in dataloader:
            self.model.eval()
            with torch.no_grad():
                inputs = SmxmInput(*batch)
                outputs = self.model(**inputs)
                probability = (
                    torch.nn.Softmax(dim=-1)(outputs).cpu().numpy().tolist()
                )
                probabilities += [p for p in probability]
                outputs = torch.argmax(outputs, dim=2)
                preds += outputs.detach().cpu().numpy().tolist()

        span_annotations = predictions_to_span_annotations(
            sentences, preds, probabilities, entity_labels, self.tokenizer, max_sentence_tokens
        )

        return span_annotations



import random

import torch
from transformers import BertModel, BertPreTrainedModel, logging

logging.set_verbosity_error()

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class BertTaggerMultiClass(BertPreTrainedModel):
    def __init__(self, config):

        super().__init__(config)

        self.bert = BertModel(config)
        self.drop = torch.nn.Dropout(config.finetuning_task["dropout_prob"])
        self.bert_output_size = config.hidden_size
        self.linear = torch.nn.Linear(self.bert_output_size, 1)
        self.linear_zero2 = torch.nn.Linear(self.bert_output_size, 1)
        self.linear_zero3 = torch.nn.Linear(self.bert_output_size, 1)

        self.init_weights()

    def forward(
        self,
        *args,
        input_ids,
        attention_mask,
        token_type_ids,
        sep_index,
        seq_mask,
        split,
        **kwargs,
    ):
        sep_index_max = torch.max(sep_index)
        predictions = []
        predictions_zero = []
        predictions_zero_base = []
        for j in range(input_ids.size(0)):
            if j == 0:
                inp_zero = torch.stack(
                    [
                        input_ids[j][i, : sep_index_max.item()]
                        for i in range(sep_index.size(0))
                    ]
                )
                att_zero = torch.stack(
                    [
                        attention_mask[j][i, : sep_index_max.item()]
                        for i in range(sep_index.size(0))
                    ]
                )
                tok_type_zero = torch.stack(
                    [
                        token_type_ids[j][i, : sep_index_max.item()]
                        for i in range(sep_index.size(0))
                    ]
                )
                with torch.no_grad():
                    words_out = self.bert(
                        input_ids=inp_zero,
                        attention_mask=att_zero,
                        token_type_ids=tok_type_zero,
                    )[0]

                pooled_out = self.drop(words_out)
                logits = self.linear_zero3(pooled_out)
                predictions_zero_base.append(logits)
            else:
                with torch.no_grad():
                    words_out = self.bert(
                        input_ids=input_ids[j],
                        attention_mask=attention_mask[j],
                        token_type_ids=token_type_ids[j],
                    )[0]

                words_out = torch.stack(
                    [
                        words_out[i, : sep_index_max.item(), :]
                        for i in range(words_out.size(0))
                    ]
                )
                pooled_out = self.drop(words_out)
                predictions_zero.append(self.linear_zero2(pooled_out))
                logits = self.linear(pooled_out)
                predictions.append(logits)

        random.shuffle(predictions_zero)
        predictions_zero = torch.stack(predictions_zero_base + predictions_zero)
        predictions_zero = predictions_zero.transpose(0, 1).transpose(1, 2)
        predictions_zero = predictions_zero.contiguous().view(
            predictions_zero.size(0), predictions_zero.size(1), -1
        )

        predictions_zero = torch.max(predictions_zero, dim=2)[0].unsqueeze(2)

        predictions = torch.stack(predictions)
        predictions = predictions.transpose(0, 1).transpose(1, 2).squeeze(3)

        logits = torch.cat((predictions_zero, predictions), dim=2)

        return logits



import functools
import logging
import os
import pathlib
import shutil
from typing import List
from urllib.request import urlopen

import requests
from spacy.tokens import Doc
from tqdm.auto import tqdm

from spacy.util import filter_spans as spacy_filter_spans

from zshot.utils.data_models import Span


def download_file(url, output_dir="."):
    """
    Utility for downloading a file
    :param url: the file url
    :param output_dir: the output dir
    :return:
    """
    filename = url.rsplit('/', 1)[1]
    path = pathlib.Path(os.path.join(output_dir, filename)).resolve()
    path.parent.mkdir(parents=True, exist_ok=True)
    with requests.get(url, stream=True) as r:
        logging.info(f"Downloading {url}")
        total_length = int(urlopen(url=url).info().get('Content-Length', 0))
        if path.exists() and os.path.getsize(path) == total_length:
            return
        r.raw.read = functools.partial(r.raw.read, decode_content=True)
        with tqdm.wrapattr(r.raw, "read", total=total_length, desc=f"Downloading {filename}") as raw:
            with path.open("wb") as output:
                shutil.copyfileobj(raw, output)


def filter_extended_spans(spans: List[Span], doc: Doc = None) -> List[Span]:
    """ If a token belongs to more than one entity, will return only the one with the higher confidence.

    :param spans: List of spans to filter
    :param doc: The Spacy Document
    :return: List of spans without overlap
    """
    spacy_spans = [span.to_spacy_span(doc) for span in spans]
    spans = [Span.from_spacy_span(spacy_span, score=span.score) for spacy_span, span in zip(spacy_spans, spans)]
    if not all([s.score for s in spans]):
        spans = spacy_filter_spans(spacy_spans)
        spans = [Span.from_spacy_span(span) for span in spans]
        return spans

    # Find overlaps
    rs = []
    spans_grouped = []
    for i, span in enumerate(spans):
        r = range(span.start, span.end + 1)
        group = [span]
        if any([set(r) & set(r_) for r_ in rs]):
            continue

        for span_ in spans[i:]:
            r_ = range(span_.start, span_.end + 1)
            if set(r) & set(r_):
                r = range(min(r[0], r_[0]), max(r[-1], r_[-1]))
                group.append(span_)

        rs.append(r)
        spans_grouped.append(group)

    # Fix overlaps getting the one with higher score
    final_mentions = []
    for group in spans_grouped:
        final_mentions.append(max(group, key=lambda x: x.score))

    return final_mentions



from pydoc import Doc
from typing import Dict

from spacy import displacy as s_displacy


def ents_colors(doc: Doc):
    """
    Can be used to derive colors for entities in a Spacy document.
    A color for each entity type in generated, using the entity label hash
    :param doc: A Spacy document with entities
    :return: A colors dictionary containing a color for each entity type
    """

    def color_from_label(label: str):
        hash_s = hash(label)
        r = (hash_s & 0xFF0000) >> 16
        g = (hash_s & 0x00FF00) >> 8
        b = hash_s & 0x0000FF
        return '#%02x%02x%02x' % (r, g, b)

    labels = set(ent.label_ for ent in doc.ents)
    colors = dict([(ent, color_from_label(ent)) for ent in labels])
    return colors


class displacy:

    @staticmethod
    def render(doc, options: Dict = None, **kwargs) -> str:
        if options:
            options['colors'] = ents_colors(doc)
        else:
            options = {'colors': ents_colors(doc)}
        return s_displacy.render(doc, options=options, **kwargs)

    @staticmethod
    def serve(doc, options: Dict = None, **kwargs):
        if options:
            options['colors'] = ents_colors(doc)
        else:
            options = {'colors': ents_colors(doc)}
        s_displacy.serve(doc, options=options, **kwargs)



from typing import Dict, List, Union

from datasets import Dataset
from evaluate import TokenClassificationEvaluator

from zshot.utils.alignment_utils import filter_overlapping_spans, AlignmentMode
from zshot.utils.data_models import Span


class ZeroShotTokenClassificationEvaluator(TokenClassificationEvaluator):

    def __init__(self, task="token-classification", default_metric_name=None, alignment_mode=AlignmentMode.expand):
        super().__init__(task, default_metric_name)
        self.alignment_mode = alignment_mode

    def predictions_processor(self, predictions: List[List[Dict]], sentences: List[List[str]], join_by: str):
        predictions_pr = []
        for prediction, words in zip(predictions, sentences):
            words_offsets = self.words_to_offsets(words, join_by)
            prediction_spans = list(map(lambda p: Span(p['start'], p['end'] - 1, p['entity'],
                                                       p['score'] if 'score' in p else None), prediction))
            filter_dict = filter_overlapping_spans(prediction_spans,
                                                   words,
                                                   tokens_offsets=words_offsets,
                                                   alignment_mode=self.alignment_mode,
                                                   return_dict=True)
            predictions_pr.append(filter_dict['bio'])
        return {"predictions": predictions_pr}

    def prepare_pipeline(
            self,
            model_or_pipeline,  # noqa: F821
            tokenizer=None,  # noqa: F821
            feature_extractor=None,  # noqa: F821
            device: int = None,
    ):
        pipe = super(TokenClassificationEvaluator, self).prepare_pipeline(model_or_pipeline, tokenizer, feature_extractor, device)
        return pipe


class MentionsExtractorEvaluator(ZeroShotTokenClassificationEvaluator):
    def parse_label(self, label):
        if label != "O":
            if label.startswith("B-") or label.startswith("I-"):
                label = label[:2] + "MENTION"
            else:
                label = "MENTION"

        return label

    def prepare_data(self, data: Union[str, Dataset], input_column: str, label_column: str, join_by: str):
        metric_inputs, pipeline_inputs = super().prepare_data(data, input_column, label_column, join_by)

        metric_inputs['references'] = [[self.parse_label(label) for label in sent]
                                       for sent in metric_inputs['references']]

        return metric_inputs, pipeline_inputs



import json
import os
import tempfile
from typing import List

import spacy

from zshot import Zshot, PipelineConfig
from zshot.utils.data_models import Entity
from zshot.tests.config import EX_DOCS, EX_ENTITIES
from zshot.tests.linker.test_linker import DummyLinker, DummyLinkerEnd2End
from zshot.tests.mentions_extractor.test_mention_extractor import DummyMentionsExtractor, DummyMentionsExtractorWithNER


def test_add_pipe():
    nlp = spacy.blank("en")
    nlp.add_pipe("zshot")
    assert "zshot" in nlp.pipe_names


def test_disable_ner():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("zshot", last=True)
    assert "zshot" in nlp.pipe_names
    assert "ner" not in nlp.pipe_names


def test_disable_mentions_extractor():
    nlp = spacy.load("en_core_web_sm")
    config_zshot = PipelineConfig(mentions_extractor=DummyMentionsExtractorWithNER(), linker=DummyLinkerEnd2End())
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names and "ner" not in nlp.pipe_names
    assert not nlp.get_pipe("zshot").mentions_extractor


def test_serialization_zshot():
    nlp = spacy.blank("en")
    nlp.add_pipe("zshot", last=True)
    assert "zshot" in nlp.pipe_names
    assert "ner" not in nlp.pipe_names
    pipes = [p for p in nlp.pipe_names if p != "zshot"]

    d = tempfile.TemporaryDirectory()
    nlp.to_disk(d.name, exclude=pipes)
    config_fn = os.path.join(d.name, "zshot", "config.cfg")
    assert os.path.exists(config_fn)
    with open(config_fn, "r") as f:
        config = json.load(f)
    assert "disable_default_ner" in config and config["disable_default_ner"]
    nlp2 = spacy.load(d.name)
    assert "zshot" in nlp2.pipe_names


def test_call_pipe_with_registered_function_configuration():

    @spacy.registry.misc("dummy.mentions-extractor")
    def create_custom_spacy_extractor():
        return DummyMentionsExtractor()

    @spacy.registry.misc("dummy.linker")
    def create_custom_linker():
        return DummyLinker()

    @spacy.registry.misc("get.entities.v1")
    def get_entities() -> List[Entity]:
        return EX_ENTITIES

    config_zshot = {
        "mentions_extractor": {"@misc": "dummy.mentions-extractor"},
        "linker": {"@misc": "dummy.linker"},
        "entities": {"@misc": "get.entities.v1"}
    }

    nlp = spacy.blank("en")
    nlp.add_pipe("zshot", config=config_zshot, last=True)
    assert "zshot" in nlp.pipe_names
    zshot_component: Zshot = [comp for name, comp in nlp.pipeline if name == 'zshot'][0]
    assert len(zshot_component.entities) == len(EX_ENTITIES)
    assert type(zshot_component.entities[0]) == Entity


def test_call_pipe_with_piepeline_configuration():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("zshot", config=PipelineConfig(
        mentions_extractor=DummyMentionsExtractor(),
        linker=DummyLinker(),
        entities=EX_ENTITIES), last=True)
    assert "zshot" in nlp.pipe_names
    zshot_component: Zshot = [comp for name, comp in nlp.pipeline if name == 'zshot'][0]
    assert len(zshot_component.entities) == len(EX_ENTITIES)
    assert type(zshot_component.entities[0]) == Entity


def test_process_single_document():
    nlp = spacy.blank("en")
    nlp.add_pipe("zshot", last=True)
    assert "zshot" in nlp.pipe_names
    doc = nlp(EX_DOCS[0])
    assert doc._.mentions is not None


def test_process_pipeline_documents():
    nlp = spacy.blank("en")
    nlp.add_pipe("zshot", last=True)
    assert "zshot" in nlp.pipe_names
    assert all(doc._.mentions is not None for doc in nlp.pipe(EX_DOCS))



from enum import Enum
from itertools import accumulate
from operator import attrgetter
from typing import List, Union, Dict, Tuple

from spacy.tokens import Doc

from zshot.utils.data_models import Span


class AlignmentMode(str, Enum):
    expand = 'expand'
    contract = 'contract'


def spacy_token_offsets(doc: Doc) -> List[Tuple[int, int]]:
    return [(t.idx, t.idx + len(t.text)) for t in doc]


def align_spans(spans: List[Span], tokens: List[str], tokens_offsets: List[Tuple[int, int]] = None,
                join_by: str = None, alignment_mode: AlignmentMode = AlignmentMode.expand,
                return_dict=False) -> Union[Dict, List[List[int]]]:
    """
    Align spans to a given list of tokens
    :param spans: the list of spans
    :param tokens: the tokens
    :param tokens_offsets: Tokens offset, spans of the tokens.
    Either tokens_offsets of join_by must be provided to compute spans.
    :param join_by: string used to join tokens. Either tokens_offsets of join_by must be provided to compute spans.
    :param alignment_mode: "contract" (span of all tokens completely within the character span),
     "expand" (span of all tokens at least partially covered by the character span).
    :param return_dict: If true, return alignment and tokens_offsets as Dict
    :return: alignment list of list of int where len(tokens) == len(alignment) and each alignment list are the
    index of the matching span or [] if no match is detected
    """
    assert join_by is not None or tokens_offsets is not None, \
        "Either tokens_offsets of join_by must be provided to compute spans."
    if not tokens_offsets:
        tokens_map = list(accumulate(map(lambda t: len(t) + len(join_by), tokens)))
        tokens_offsets = list(zip([0] + tokens_map, map(lambda x: x - len(join_by), tokens_map)))
    alignments = [[] for _ in range(len(tokens))]
    for idt, (t_start, t_end) in enumerate(tokens_offsets):
        for ids, s in enumerate(spans):
            if (t_start <= s.start < t_end or t_start < s.end <= t_end) and alignment_mode == AlignmentMode.expand:
                alignments[idt].append(ids)
            if t_start >= s.start and t_end <= s.end and alignment_mode == AlignmentMode.contract:
                alignments[idt].append(ids)
    if return_dict:
        return {
            'tokens_offsets': tokens_offsets,
            'alignments': alignments,
        }
    return alignments


def filter_overlapping_spans(spans: List[Span], tokens: List[str],
                             tokens_offsets: List[Tuple[int, int]] = None,
                             join_by: str = None,
                             alignment_mode: AlignmentMode = AlignmentMode.expand,
                             return_dict=False) -> Union[List[Span], Dict]:
    """
    :param spans: List of spans to align
    :param tokens: List of tokens
    :param tokens_offsets: Tokens offset, spans of the tokens
    :param join_by: string used to join tokens. Either tokens_offsets of join_by must be provided to compute spans.
    :param return_dict: If true, return filtered list of spans and partial results as Dict
    :param alignment_mode: "contract" (span of all tokens completely within the character span),
     "expand" (span of all tokens at least partially covered by the character span).
    :return: the filtered list of spans
    """
    align_dict = align_spans(spans, tokens, tokens_offsets, join_by=join_by,
                             alignment_mode=alignment_mode, return_dict=True)
    alignments = align_dict['alignments']
    tokens_offsets = align_dict['tokens_offsets']
    filtered_spans = [None] * len(alignments)
    bio_token = ['O'] * len(alignments)
    for idx, alignment in enumerate(alignments):
        t_spans = [spans[a] for a in alignment]
        if not t_spans:
            continue
        try:
            best_span = max(t_spans, key=attrgetter('score'))
        except TypeError:
            best_span = t_spans[0]
        if idx > 0 and filtered_spans[idx - 1] is not None and filtered_spans[idx - 1].label == best_span.label:
            filtered_spans[idx - 1].end = tokens_offsets[idx][1]
            bio_token[idx] = f"I-{filtered_spans[idx - 1].label}"
        else:
            filtered_spans[idx] = Span(start=tokens_offsets[idx][0], end=tokens_offsets[idx][1],
                                       label=best_span.label, kb_id=best_span.kb_id)
            bio_token[idx] = f"B-{filtered_spans[idx].label}"

    filtered_spans = list(filter(None, filtered_spans))
    if return_dict:
        return {
            'bio': bio_token,
            'filtered_spans': filtered_spans,
            'alignments': alignments,
            'tokens_offsets': tokens_offsets
        }
    return filtered_spans
